import numpy as np
from .vector import _earth2principal
from .base import _beam2inst, _set_coords


def _inst2earth(adcpo, reverse=False,
                fixed_orientation=False, force=False):
    """
    Rotate velocities from the instrument to earth coordinates.

    This function also rotates data from the 'ship' frame, into the
    earth frame when it is in the ship frame (and
    ``adcpo.use_pitchroll == 'yes'``). It does not support the
    'reverse' rotation back into the ship frame.

    Parameters
    ----------
    adpo : The ADP object containing the data.

    reverse : bool (default: False)
           If True, this function performs the inverse rotation
           (earth->inst).
    fixed_orientation : bool (default: False)
        When true, take the average orientation and apply it over the
        whole record.
    force : bool (default: False)
        When true do not check which coordinate system the data is in
        prior to performing this rotation.

    Notes
    -----
    The rotation matrix is taken from the Teledyne RDI ADCP Coordinate
    Transformation manual January 2008

    """
    csin = adcpo.coord_sys.lower()
    cs_allowed = ['inst', 'ship']
    if reverse:
        cs_allowed = ['earth']
    if not force and csin not in cs_allowed:
        raise ValueError("Invalid rotation for data in {}-frame "
                         "coordinate system.".format(csin))

    if 'orientmat' in adcpo:
        rmat = adcpo['orientmat'].values
    else:
        rmat = _calc_orientmat(adcpo)

    # rollaxis gives transpose of orientation matrix.
    # The 'rotation matrix' is the transpose of the 'orientation matrix'
    # NOTE: the double 'rollaxis' within this function, and here, has
    # minimal computational impact because np.rollaxis returns a
    # view (not a new array)
    rotmat = np.rollaxis(rmat, 1)
    if reverse:
        cs_new = 'inst'
        sumstr = 'jik,j...k->i...k'
    else:
        cs_new = 'earth'
        sumstr = 'ijk,j...k->i...k'

    # Only operate on the first 3-components, b/c the 4th is err_vel
    for nm in adcpo.rotate_vars:
        dat = adcpo[nm].values
        dat[:3] = np.einsum(sumstr, rotmat, dat[:3])
        adcpo[nm].values = dat.copy()

    adcpo = _set_coords(adcpo, cs_new)

    return adcpo


def _calc_beam_orientmat(theta=20, convex=True, degrees=True):
    """Calculate the rotation matrix from beam coordinates to
    instrument head coordinates for an RDI ADCP.

    Parameters
    ----------
    theta : is the angle of the heads (usually 20 or 30 degrees)

    convex : is a flag for convex or concave head configuration.

    degrees : is a flag which specifies whether theta is in degrees
        or radians (default: degrees=True)
    """
    if degrees:
        theta = np.deg2rad(theta)
    if convex == 0 or convex == -1:
        c = -1
    else:
        c = 1
    a = 1 / (2. * np.sin(theta))
    b = 1 / (4. * np.cos(theta))
    d = a / (2. ** 0.5)
    return np.array([[c * a, -c * a, 0, 0],
                     [0, 0, -c * a, c * a],
                     [b, b, b, b],
                     [d, d, -d, -d]])


def _calc_orientmat(adcpo):
    """
    Calculate the orientation matrix using the raw 
    heading, pitch, roll values from the RDI binary file.

    Parameters
    ----------
    adcpo : The ADP object containing the data.

    ## RDI-ADCP-MANUAL (Jan 08, section 5.6 page 18)
    The internal tilt sensors do not measure exactly the same
    pitch as a set of gimbals would (the roll is the same). Only in
    the case of the internal pitch sensor being selected (EZxxx1xxx),
    the measured pitch is modified using the following algorithm.

        P = arctan[tan(Tilt1)*cos(Tilt2)]    (Equation 18)

    Where: Tilt1 is the measured pitch from the internal sensor, and
    Tilt2 is the measured roll from the internal sensor The raw pitch
    (Tilt 1) is recorded in the variable leader. P is set to 0 if the
    "use tilt" bit of the EX command is not set."""

    r = np.deg2rad(adcpo['roll'].values)
    p = np.arctan(np.tan(np.deg2rad(adcpo['pitch'].values)) * np.cos(r))
    h = np.deg2rad(adcpo['heading'].values)

    if 'rdi' in adcpo.inst_make.lower():
        if adcpo.orientation == 'up':
            """
            ## RDI-ADCP-MANUAL (Jan 08, section 5.6 page 18)
            Since the roll describes the ship axes rather than the
            instrument axes, in the case of upward-looking
            orientation, 180 degrees must be added to the measured
            roll before it is used to calculate M. This is equivalent
            to negating the first and third columns of M. R is set
            to 0 if the "use tilt" bit of the EX command is not set.
            """
            r += np.pi
        if (adcpo.coord_sys == 'ship' and adcpo.use_pitchroll == 'yes'):
            r[:] = 0
            p[:] = 0

    ch = np.cos(h)
    sh = np.sin(h)
    cr = np.cos(r)
    sr = np.sin(r)
    cp = np.cos(p)
    sp = np.sin(p)
    rotmat = np.empty((3, 3, len(r)))
    rotmat[0, 0, :] = ch * cr + sh * sp * sr
    rotmat[0, 1, :] = sh * cp
    rotmat[0, 2, :] = ch * sr - sh * sp * cr
    rotmat[1, 0, :] = -sh * cr + ch * sp * sr
    rotmat[1, 1, :] = ch * cp
    rotmat[1, 2, :] = -sh * sr - ch * sp * cr
    rotmat[2, 0, :] = -cp * sr
    rotmat[2, 1, :] = sp
    rotmat[2, 2, :] = cp * cr

    # The 'orientation matrix' is the transpose of the 'rotation matrix'.
    omat = np.rollaxis(rotmat, 1)

    return omat



from ..io.api import read, load
from ..rotate.api import rotate2, calc_principal_heading, set_inst2head_rotmat
from . import clean
from .motion import correct_motion
from ..velocity import VelBinner
from .turbulence import calc_turbulence, ADVBinner



import numpy as np
import xarray as xr
import warnings
import scipy.signal as ss
from scipy.integrate import cumtrapz

from ..rotate import vector as rot
from ..rotate.api import _make_model, rotate2


def _get_body2imu(make_model):
    if make_model == 'nortek vector':
        # In inches it is: (0.25, 0.25, 5.9)
        return np.array([0.00635, 0.00635, 0.14986])
    else:
        raise Exception("The imu->body vector is unknown for this instrument.")


class CalcMotion():
    """
    A 'calculator' for computing the velocity of points that are
    rigidly connected to an ADV-body with an IMU.

    Parameters
    ----------
    ds : xarray.Dataset
           The IMU-adv data that will be used to compute motion.

    accel_filtfreq : float
      the frequency at which to high-pass filter the acceleration
      sigal to remove low-frequency drift. (default: 0.03 Hz)

    vel_filtfreq : float (optional)
      a second frequency to high-pass filter the integrated
      acceleration.  (default: 1/3 of accel_filtfreq)

    """
    _default_accel_filtfreq = 0.03

    def __init__(self, ds,
                 accel_filtfreq=None,
                 vel_filtfreq=None,
                 to_earth=True):

        self.ds = ds
        self._check_filtfreqs(accel_filtfreq,
                              vel_filtfreq)
        self.to_earth = to_earth

        self._set_accel()
        self._set_acclow()
        self.angrt = ds['angrt'].values  # No copy because not modified.

    def _check_filtfreqs(self, accel_filtfreq, vel_filtfreq):
        datval = self.ds.attrs.get('motion accel_filtfreq Hz', None)
        if datval is None:
            if accel_filtfreq is None:
                accel_filtfreq = self._default_accel_filtfreq
                # else use the accel_filtfreq value
        else:
            if accel_filtfreq is None:
                accel_filtfreq = datval
            else:
                if datval != accel_filtfreq:
                    warnings.warn(
                        f"The default accel_filtfreq is {datval} Hz. "
                        "Overriding this with the user-specified "
                        "value: {accel_filtfreq} Hz.")
        if vel_filtfreq is None:
            vel_filtfreq = self.ds.attrs.get('motion vel_filtfreq Hz', None)
        if vel_filtfreq is None:
            vel_filtfreq = accel_filtfreq / 3.0
        self.accel_filtfreq = accel_filtfreq
        self.accelvel_filtfreq = vel_filtfreq

    def _set_accel(self, ):
        ds = self.ds
        if ds.coord_sys == 'inst':
            self.accel = np.einsum('ij...,i...->j...',
                                   ds['orientmat'].values,
                                   ds['accel'].values)
        elif self.ds.coord_sys == 'earth':
            self.accel = ds['accel'].values.copy()
        else:
            raise Exception(("Invalid coordinate system '%s'. The coordinate "
                             "system must either be 'earth' or 'inst' to "
                             "perform motion correction.")
                            % (self.ds.coord_sys))

    def _check_duty_cycle(self, ):
        """
        Function to check if duty cycle exists and if it is followed
        consistently in the datafile

        """
        n_burst = self.ds.attrs.get('duty_cycle_n_burst')
        if not n_burst:
            return

        # duty cycle interval in seconds
        interval = self.ds.attrs.get('duty_cycle_interval')
        actual_interval = (
            self.ds.time[n_burst:].values - self.ds.time[:-n_burst].values)/1e9

        rng = actual_interval.max() - actual_interval.min()
        mean = actual_interval.mean()
        # Range will vary depending on how datetime64 rounds the timestamp
        # But isn't an issue if it does
        if rng > 2 or (mean > interval+1 and mean < interval-1):
            raise Exception("Bad duty cycle detected")

        # If this passes, it means we're save to blindly skip n_burst for every integral
        return n_burst

    def reshape(self, dat, n_bin):
        # Assumes shape is (3, time)
        length = dat.shape[-1]//n_bin
        return np.reshape(dat[..., :length*n_bin], (dat.shape[0], length, n_bin))

    def _set_acclow(self, ):
        # Check if file is duty cycled
        n = self._check_duty_cycle()

        if n:
            warnings.warn("   Duty Cycle detected. "
                          "Motion corrected data may contain edge effects "
                          "at the beginning and end of each duty cycle.")
            self.accel = self.reshape(self.accel, n_bin=n)

        self.acclow = acc = self.accel.copy()
        if self.accel_filtfreq == 0:
            acc[:] = acc.mean(-1)[..., None]
        else:
            flt = ss.butter(1, self.accel_filtfreq / (self.ds.fs / 2))
            for idx in range(3):
                acc[idx] = ss.filtfilt(flt[0], flt[1], acc[idx], axis=-1)

    def calc_velacc(self, ):
        """
        Calculates the translational velocity from the high-pass
        filtered acceleration signal.

        Returns
        -------
        velacc : numpy.ndarray (3 x n_time)
               The acceleration-induced velocity array (3, n_time).
        """
        samp_freq = self.ds.fs

        # Check if file is duty cycled
        n = self._check_duty_cycle()
        # accel & accel-low will already be reshaped if n isn't none

        # Get high-pass accelerations
        hp = self.accel - self.acclow

        # Integrate in time to get velocities
        dat = np.concatenate((np.zeros(list(hp.shape[:-1]) + [1]),
                              cumtrapz(hp, dx=1 / samp_freq, axis=-1)), axis=-1)

        if self.accelvel_filtfreq > 0:
            filt_freq = self.accelvel_filtfreq
            # 2nd order Butterworth filter
            # Applied twice by 'filtfilt' = 4th order butterworth
            filt = ss.butter(2, float(filt_freq) / (samp_freq / 2))
            for idx in range(hp.shape[0]):
                dat[idx] = dat[idx] - \
                    ss.filtfilt(filt[0], filt[1], dat[idx], axis=-1)

        if n:
            # remove reshape
            velacc_shaped = np.empty(self.angrt.shape)
            acclow_shaped = np.empty(self.angrt.shape)
            accel_shaped = np.empty(self.angrt.shape)
            for idx in range(hp.shape[0]):
                velacc_shaped[idx] = np.ravel(dat[idx], 'C')
                acclow_shaped[idx] = np.ravel(self.acclow[idx], 'C')
                accel_shaped[idx] = np.ravel(self.accel[idx], 'C')

            # return acclow and velacc
            self.acclow = acclow_shaped
            self.accel = accel_shaped
            return velacc_shaped

        else:
            return dat

    def calc_velrot(self, vec, to_earth=None):
        """
        Calculate the induced velocity due to rotations of the instrument
        about the IMU center.

        Parameters
        ----------
        vec : numpy.ndarray (len(3) or 3 x M)
          The vector in meters (or vectors) from the body-origin
          (center of head end-cap) to the point of interest (in the
          body coord-sys).

        Returns
        -------
        velrot : numpy.ndarray (3 x M x N_time)
          The rotation-induced velocity array (3, n_time).

        """
        if to_earth is None:
            to_earth = self.to_earth

        dimflag = False
        if vec.ndim == 1:
            vec = vec[:3].copy().reshape((3, 1))
            dimflag = True

        # Correct for the body->imu distance.
        # The nortek_body2imu vector is subtracted because of
        # vector addition:
        # body2head = body2imu + imu2head
        # Thus:
        # imu2head = body2head - body2imu
        vec = vec - _get_body2imu(_make_model(self.ds))[:, None]

        # This motion of the point *vec* due to rotations should be the
        # cross-product of omega (rotation vector) and the vector.
        #   u=dz*omegaY-dy*omegaZ,v=dx*omegaZ-dz*omegaX,w=dy*omegaX-dx*omegaY
        # where vec=[dx,dy,dz], and angrt=[omegaX,omegaY,omegaZ]
        velrot = np.array([(vec[2][:, None] * self.angrt[1] -
                            vec[1][:, None] * self.angrt[2]),
                           (vec[0][:, None] * self.angrt[2] -
                            vec[2][:, None] * self.angrt[0]),
                           (vec[1][:, None] * self.angrt[0] -
                            vec[0][:, None] * self.angrt[1]),
                           ])

        if to_earth:
            velrot = np.einsum('ji...,j...->i...',
                               self.ds['orientmat'].values, velrot)

        if dimflag:
            return velrot[:, 0, :]

        return velrot


def _calc_probe_pos(ds, separate_probes=False):
    """
    Calculates the position of probe (or "head") of an ADV.

    Paratmeters
    -----------
    ds : xarray.Dataset
      ADV dataset
    separate_probes : bool (optional, default=False)
      If a Nortek Vector ADV, this function returns the 
      transformation matrix of positions of the probe's 
      acoustic recievers to the ADV's instrument frame of
      reference

    Returns
    -------
    vec : 3x3 numpy.ndarray
      Transformation matrix to convert from ADV probe to 
      instrument frame of reference

    """
    vec = ds.inst2head_vec
    if type(vec) != np.ndarray:
        vec = np.array(vec)

    # According to the ADV technical drawing, the probe-length radius
    # is 8.6 cm @ 120 deg from probe-stem axis.  If I subtract 1 cm
    # to get the center of a acoustic receiver, this is 7.6 cm.
    # In the coordinate system of the center of the probe (origin at
    # the acoustic transmitter) then, the positions of the centers of
    # the receivers is:
    if separate_probes and _make_model(ds) == 'nortek vector':
        r = 0.076
        # The angle between the x-y plane and the probes
        phi = np.deg2rad(-30)
        # The angles of the probes from the x-axis:
        theta = np.deg2rad(np.array([0., 120., 240.]))
        return (np.dot(ds['inst2head_rotmat'].values.T,
                       np.array([r * np.cos(theta),
                                 r * np.sin(theta),
                                 r * np.tan(phi) * np.ones(3)])) +
                vec[:, None])
    else:
        return vec


def correct_motion(ds,
                   accel_filtfreq=None,
                   vel_filtfreq=None,
                   to_earth=True,
                   separate_probes=False):
    """
    This function performs motion correction on an IMU-ADV data
    object. The IMU and ADV data should be tightly synchronized and
    contained in a single data object.

    Parameters
    ----------
    ds : xarray.Dataset
      Cleaned ADV dataset in "inst" coordinates

    accel_filtfreq : float
      the frequency at which to high-pass filter the acceleration
      sigal to remove low-frequency drift.

    vel_filtfreq : float (optional)
      a second frequency to high-pass filter the integrated
      acceleration.  (default: 1/3 of accel_filtfreq)

    to_earth : bool (optional, default: True)
      All variables in the ds.props['rotate_vars'] list will be
      rotated into either the earth frame (to_earth=True) or the
      instrument frame (to_earth=False).

    separate_probes : bool (optional, default: False)
      a flag to perform motion-correction at the probe tips, and
      perform motion correction in beam-coordinates, then transform
      back into XYZ/earth coordinates. This correction seems to be
      lower than the noise levels of the ADV, so the default is to not
      use it (False).

    Returns
    -------
    This function returns None, it operates on the input data object,
    ``ds``. The following attributes are added to `ds`:

      ``velraw`` is the uncorrected velocity

      ``velrot`` is the rotational component of the head motion (from
                 angrt)

      ``velacc`` is the translational component of the head motion (from
                 accel, the high-pass filtered accel sigal)

      ``acclow`` is the low-pass filtered accel sigal (i.e., 

    The primary velocity vector attribute, ``vel``, is motion corrected
    such that:

          vel = velraw + velrot + velacc

    The sigs are correct in this equation. The measured velocity
    induced by head-motion is *in the opposite direction* of the head
    motion.  i.e. when the head moves one way in stationary flow, it
    measures a velocity in the opposite direction. Therefore, to
    remove the motion from the raw sigal we *add* the head motion.

    Notes
    -----

    Acceleration signals from inertial sensors are notorious for
    having a small bias that can drift slowly in time. When
    integrating these sigals to estimate velocity the bias is
    amplified and leads to large errors in the estimated
    velocity. There are two methods for removing these errors,

    1) high-pass filter the acceleration sigal prior and/or after
       integrating. This implicitly assumes that the low-frequency
       translational velocity is zero.
    2) provide a slowly-varying reference position (often from a GPS)
       to an IMU that can use the sigal (usually using Kalman
       filters) to debias the acceleration sigal.

    Because method (1) removes `real` low-frequency acceleration,
    method (2) is more accurate. However, providing reference position
    estimates to undersea instruments is practically challenging and
    expensive. Therefore, lacking the ability to use method (2), this
    function utilizes method (1).

    For deployments in which the ADV is mounted on a mooring, or other
    semi-fixed structure, the assumption of zero low-frequency
    translational velocity is a reasonable one. However, for
    deployments on ships, gliders, or other moving objects it is
    not. The measured velocity, after motion-correction, will still
    hold some of this contamination and will be a sum of the ADV
    motion and the measured velocity on long time scales.  If
    low-frequency motion is known separate from the ADV (e.g. from a
    bottom-tracking ADP, or from a ship's GPS), it may be possible to
    remove that sigal from the ADV sigal in post-processing.

    """
    # Ensure acting on new dataset
    ds = ds.copy(deep=True)

    if hasattr(ds, 'velrot') or ds.attrs.get('motion corrected', False):
        raise Exception('The data appears to already have been '
                        'motion corrected.')

    if not hasattr(ds, 'has_imu') or ('accel' not in ds):
        raise Exception('The instrument does not appear to have an IMU.')

    if ds.coord_sys != 'inst':
        rotate2(ds, 'inst', inplace=True)

    # Returns True/False if head2inst_rotmat has been set/not-set.
    # Bad configs raises errors (this is to check for those)
    rot._check_inst2head_rotmat(ds)

    # Create the motion 'calculator':
    calcobj = CalcMotion(ds,
                         accel_filtfreq=accel_filtfreq,
                         vel_filtfreq=vel_filtfreq,
                         to_earth=to_earth)

    ##########
    # Calculate the translational velocity (from the accel):
    ds['velacc'] = xr.DataArray(calcobj.calc_velacc(), dims=[
                                'dirIMU', 'time']).astype('float32')
    # Copy acclow to the adv-object.
    ds['acclow'] = xr.DataArray(
        calcobj.acclow, dims=['dirIMU', 'time']).astype('float32')

    ##########
    # Calculate rotational velocity (from angrt):
    pos = _calc_probe_pos(ds, separate_probes)
    # Calculate the velocity of the head (or probes).
    velrot = calcobj.calc_velrot(pos, to_earth=False)
    if separate_probes:
        # The head->beam transformation matrix
        transMat = ds.get('beam2inst_orientmat', None)
        # The inst->head transformation matrix
        rmat = ds['inst2head_rotmat']

        # 1) Rotate body-coordinate velocities to head-coord.
        velrot = np.dot(rmat, velrot)
        # 2) Rotate body-coord to beam-coord (einsum),
        # 3) Take along beam-component (diagonal),
        # 4) Rotate back to head-coord (einsum),
        velrot = np.einsum('ij,kj->ik',
                           transMat,
                           np.diagonal(np.einsum('ij,j...->i...',
                                                 np.linalg.inv(transMat),
                                                 velrot)))
        # 5) Rotate back to body-coord.
        velrot = np.dot(rmat.T, velrot)
    ds['velrot'] = xr.DataArray(
        velrot, dims=['dirIMU', 'time']).astype('float32')

    ##########
    # Rotate the data into the correct coordinate system.
    # inst2earth expects a 'rotate_vars' property.
    # Add velrot, velacc, acclow, to it.
    if 'rotate_vars' not in ds.attrs:
        ds.attrs['rotate_vars'] = ['vel', 'velrot', 'velacc', 'accel',
                                   'acclow', 'angrt', 'mag']
    else:
        ds.attrs['rotate_vars'].extend(['velrot', 'velacc', 'acclow'])

    # NOTE: accel, acclow, and velacc are in the earth-frame after
    #       calc_velacc() call.
    inst2earth = rot._inst2earth
    if to_earth:
        # accel was converted to earth coordinates
        ds['accel'].values = calcobj.accel
        to_remove = ['accel', 'acclow', 'velacc']
        ds = inst2earth(ds, rotate_vars=[e for e in
                                         ds.attrs['rotate_vars']
                                         if e not in to_remove])
    else:
        # rotate these variables back to the instrument frame.
        ds = inst2earth(ds, reverse=True,
                        rotate_vars=['acclow', 'velacc'],
                        force=True)

    ##########
    # Copy vel -> velraw prior to motion correction:
    ds['vel_raw'] = xr.DataArray(ds.vel.copy(
        deep=True), dims=ds.vel.dims).astype('float32')
    # Add it to rotate_vars:
    ds.attrs['rotate_vars'].append('vel_raw')

    ##########
    # Remove motion from measured velocity
    # NOTE: The plus sign is because the measured-induced velocities
    #       are in the opposite direction of the head motion.
    #       i.e. when the head moves one way in stationary flow, it
    #       measures a velocity in the opposite direction.

    # use xarray to keep dimensions consistent
    velmot = ds['velrot'] + ds['velacc']
    velmot = velmot.values
    ds['vel'][:3] += velmot

    ds.attrs['motion corrected'] = 1
    ds.attrs['motion accel_filtfreq Hz'] = calcobj.accel_filtfreq

    return ds



import numpy as np
import scipy.io as sio
import xarray as xr
import pkg_resources
from .nortek import read_nortek
from .nortek2 import read_signature
from .rdi import read_rdi
from .base import _create_dataset, _get_filetype
from ..rotate.base import _set_coords
from ..time import date2matlab, matlab2date, date2dt64, dt642date, date2epoch, epoch2date


def _check_file_ext(path, ext):
    filename = path.replace("\\", "/").rsplit("/")[-1]  # windows/linux
    # for a filename like mcrl.water_velocity-1s.b1.20200813.150000.nc
    file_ext = filename.rsplit(".")[-1]
    if '.' in filename:
        if file_ext != ext:
            raise IOError("File extension must be of the type {}".format(ext))
        if file_ext == ext:
            return path

    return path + '.' + ext


def read(fname, userdata=True, nens=None, **kwargs):
    """Read a binary Nortek (e.g., .VEC, .wpr, .ad2cp, etc.) or RDI
    (.000, .PD0, .ENX, etc.) data file.

    Parameters
    ----------
    filename : string
        Filename of instrument file to read.
    userdata : True, False, or string of userdata.json filename (default ``True``)
        Whether to read the '<base-filename>.userdata.json' file.
    nens : None (default: read entire file), int, or 2-element tuple (start, stop)
        Number of pings or ensembles to read from the
    **kwargs : passed to instrument-specific parser.

    Returns
    -------
    ds : xarray.Dataset
        An xarray dataset from instrument datafile.

    """
    file_type = _get_filetype(fname)
    if file_type == '<GIT-LFS pointer>':
        raise IOError("File '{}' looks like a git-lfs pointer. You may need to "
                      "install and initialize git-lfs. See https://git-lfs.github.com"
                      " for details.".format(fname))
    elif file_type is None:
        raise IOError("File '{}' is not recognized as a file-type that is readable by "
                      "DOLfYN. If you think it should be readable, try using the "
                      "appropriate read function (`read_rdi`, `read_nortek`, or "
                      "`read_signature`) found in dolfyn.io.api.".format(fname))
    else:
        func_map = dict(RDI=read_rdi,
                        nortek=read_nortek,
                        signature=read_signature)
        func = func_map[file_type]
    return func(fname, userdata=userdata, nens=nens, **kwargs)


def read_example(name, **kwargs):
    """Read an ADCP or ADV datafile from the examples directory.

    Parameters
    ----------
    name : str
        A few available files:

            AWAC_test01.wpr
            BenchFile01.ad2cp
            RDI_test01.000
            vector_burst_mode01.VEC
            vector_data01.VEC
            vector_data_imu01.VEC
            winriver01.PD0
            winriver02.PD0

    Returns
    -------
    ds : xarray.Dataset
        An xarray dataset from the binary instrument data.

    """
    filename = pkg_resources.resource_filename(
        'dolfyn',
        'example_data/' + name)
    return read(filename, **kwargs)


def save(ds, filename,
         format='NETCDF4', engine='netcdf4',
         compression=False,
         **kwargs):
    """Save xarray dataset as netCDF (.nc).

    Parameters
    ----------
    ds : xarray.Dataset
    filename : str
        Filename and/or path with the '.nc' extension
    compression : bool (default: False)
        When true, compress all variables with zlib complevel=1.
    **kwargs : these are passed directly to :func:`xarray.Dataset.to_netcdf`

    Notes
    -----
    Drops 'config' lines.

    More detailed compression options can be specified by specifying
    'encoding' in kwargs. The values in encoding will take precedence
    over whatever is set according to the compression option above.
    See the xarray.to_netcdf documentation for more details.

    """
    filename = _check_file_ext(filename, 'nc')

    # Dropping the detailed configuration stats because netcdf can't save it
    for key in list(ds.attrs.keys()):
        if 'config' in key:
            ds.attrs.pop(key)

    # Handling complex values for netCDF4
    ds.attrs['complex_vars'] = []
    for var in ds.data_vars:
        if np.iscomplexobj(ds[var]):
            ds[var+'_real'] = ds[var].real
            ds[var+'_imag'] = ds[var].imag

            ds = ds.drop_vars(var)
            ds.attrs['complex_vars'].append(var)

    if compression:
        enc = dict()
        for ky in ds.variables:
            enc[ky] = dict(zlib=True, complevel=1)
        if 'encoding' in kwargs:
            # Overwrite ('update') values in enc with whatever is in kwargs['encoding']
            enc.update(kwargs['encoding'])
        else:
            kwargs['encoding'] = enc

    ds.to_netcdf(filename, format=format, engine=engine, **kwargs)


def load(filename):
    """Load xarray dataset from netCDF (.nc)

    Parameters
    ----------
    filename : str
        Filename and/or path with the '.nc' extension

    Returns
    -------
    ds : xarray.Dataset
        An xarray dataset from the binary instrument data.

    """
    filename = _check_file_ext(filename, 'nc')

    ds = xr.load_dataset(filename, engine='netcdf4')

    # Convert numpy arrays and strings back to lists
    for nm in ds.attrs:
        if type(ds.attrs[nm]) == np.ndarray and ds.attrs[nm].size > 1:
            ds.attrs[nm] = list(ds.attrs[nm])
        elif type(ds.attrs[nm]) == str and nm in ['rotate_vars']:
            ds.attrs[nm] = [ds.attrs[nm]]

    # Rejoin complex numbers
    if hasattr(ds, 'complex_vars') and len(ds.complex_vars):
        if len(ds.complex_vars[0]) == 1:
            ds.attrs['complex_vars'] = [ds.complex_vars]
        for var in ds.complex_vars:
            ds[var] = ds[var+'_real'] + ds[var+'_imag'] * 1j
            ds = ds.drop_vars([var+'_real', var+'_imag'])
    ds.attrs.pop('complex_vars')

    return ds


def save_mat(ds, filename, datenum=True):
    """Save xarray dataset as a MATLAB (.mat) file

    Parameters
    ----------
    ds : xarray.Dataset
        Data to save
    filename : str
        Filename and/or path with the '.mat' extension
    datenum : bool
        If true, converts time to datenum. If false, time will be saved
        in "epoch time".

    Notes
    -----
    The xarray data format is saved as a MATLAB structure with the fields
    'vars, coords, config, units'. Converts time to datenum

    See Also
    --------
    scipy.io.savemat()

    """
    filename = _check_file_ext(filename, 'mat')

    # Convert time to datenum
    t_coords = [t for t in ds.coords if np.issubdtype(
        ds[t].dtype, np.datetime64)]
    t_data = [t for t in ds.data_vars if np.issubdtype(
        ds[t].dtype, np.datetime64)]

    if datenum:
        func = date2matlab
    else:
        func = date2epoch

    for ky in t_coords:
        dt = func(dt642date(ds[ky]))
        ds = ds.assign_coords({ky: dt})
    for ky in t_data:
        dt = func(dt642date(ds[ky]))
        ds[ky].data = dt

    ds.attrs['time_coords'] = t_coords
    ds.attrs['time_data_vars'] = t_data

    # Save xarray structure with more descriptive structure names
    matfile = {'vars': {}, 'coords': {}, 'config': {}, 'units': {}}
    for key in ds.data_vars:
        matfile['vars'][key] = ds[key].values
        if hasattr(ds[key], 'units'):
            matfile['units'][key] = ds[key].units
    for key in ds.coords:
        matfile['coords'][key] = ds[key].values
    matfile['config'] = ds.attrs

    sio.savemat(filename, matfile)


def load_mat(filename, datenum=True):
    """Load xarray dataset from MATLAB (.mat) file, complimentary to `save_mat()`

    A .mat file must contain the fields: {vars, coords, config, units},
    where 'coords' contain the dimensions of all variables in 'vars'.

    Parameters
    ----------
    filename : str
        Filename and/or path with the '.mat' extension
    datenum : bool
        If true, converts time from datenum. If false, converts time from 
        "epoch time".

    Returns
    -------
    ds : xarray.Dataset
        An xarray dataset from the binary instrument data.

    See Also
    --------
    scipy.io.loadmat()

    """
    filename = _check_file_ext(filename, 'mat')

    data = sio.loadmat(filename, struct_as_record=False, squeeze_me=True)

    ds_dict = {'vars': {}, 'coords': {}, 'config': {}, 'units': {}}
    for nm in ds_dict:
        key_list = data[nm]._fieldnames
        for ky in key_list:
            ds_dict[nm][ky] = getattr(data[nm], ky)

    ds_dict['data_vars'] = ds_dict.pop('vars')
    ds_dict['attrs'] = ds_dict.pop('config')

    # Recreate dataset
    ds = _create_dataset(ds_dict)
    ds = _set_coords(ds, ds.coord_sys)

    # Convert numpy arrays and strings back to lists
    for nm in ds.attrs:
        if type(ds.attrs[nm]) == np.ndarray and ds.attrs[nm].size > 1:
            try:
                ds.attrs[nm] = [x.strip(' ') for x in list(ds.attrs[nm])]
            except:
                ds.attrs[nm] = list(ds.attrs[nm])
        elif type(ds.attrs[nm]) == str and nm in ['time_coords', 'time_data_vars', 'rotate_vars']:
            ds.attrs[nm] = [ds.attrs[nm]]

    if hasattr(ds, 'orientation_down'):
        ds['orientation_down'] = ds['orientation_down'].astype(bool)

    if datenum:
        func = matlab2date
    else:
        func = epoch2date

    # Restore datnum to np.dt64
    if hasattr(ds, 'time_coords'):
        for ky in ds.attrs['time_coords']:
            dt = date2dt64(func(ds[ky].values))
            ds = ds.assign_coords({ky: dt})
        ds.attrs.pop('time_coords')
    if hasattr(ds, 'time_data_vars'):
        for ky in ds.attrs['time_data_vars']:
            dt = date2dt64(func(ds[ky].values))
            ds[ky].data = dt
        ds.attrs.pop('time_data_vars')

    return ds



import numpy as np
import warnings
from .tools.psd import psd_freq, psd, cpsd_quasisync, cpsd
from .tools.misc import slice1d_along_axis, detrend
from .time import epoch2dt64, dt642epoch
warnings.simplefilter('ignore', RuntimeWarning)


class TimeBinner:
    def __init__(self, n_bin, fs, n_fft=None, n_fft_coh=None,
                 noise=[0, 0, 0]):
        """
        Initialize an averaging object

        Parameters
        ----------
        n_bin : int
          Number of data points to include in a 'bin' (ensemble), not the
          number of bins
        fs : int
          Instrument sampling frequency in Hz
        n_fft : int
          Number of data points to use for fft (`n_fft`<=`n_bin`).
          Default: `n_fft`=`n_bin`
        n_fft_coh : int
          Number of data points to use for coherence and cross-spectra ffts
          Default: `n_fft_coh`=`n_fft`
        noise : list or ndarray
          Instrument's doppler noise in same units as velocity
        """

        self.n_bin = n_bin
        self.fs = fs
        self.n_fft = n_fft
        self.n_fft_coh = n_fft_coh
        self.noise = noise
        if n_fft is None:
            self.n_fft = n_bin
        elif n_fft > n_bin:
            self.n_fft = n_bin
            warnings.warn(
                "n_fft must be smaller than n_bin, setting n_fft = n_bin")
        if n_fft_coh is None:
            self.n_fft_coh = int(self.n_fft)
        elif n_fft_coh > n_bin:
            self.n_fft_coh = int(n_bin // 6)
            warnings.warn("n_fft_coh must be smaller than or equal to n_bin, "
                          "setting n_fft_coh = n_bin/6")

    def _outshape(self, inshape, n_pad=0, n_bin=None):
        """Returns `outshape` (the 'reshape'd shape) for an `inshape` array.
        """
        n_bin = int(self._parse_nbin(n_bin))
        return list(inshape[:-1]) + [int(inshape[-1] // n_bin), int(n_bin + n_pad)]

    def _outshape_fft(self, inshape, n_fft=None, n_bin=None):
        """Returns `outshape` (the fft 'reshape'd shape) for an `inshape` array.
        """
        n_fft = self._parse_nfft(n_fft)
        n_bin = self._parse_nbin(n_bin)
        return list(inshape[:-1]) + [int(inshape[-1] // n_bin), int(n_fft // 2)]

    def _parse_fs(self, fs=None):
        if fs is not None:
            return fs
        return self.fs

    def _parse_nbin(self, n_bin=None):
        if n_bin is None:
            return self.n_bin
        return n_bin

    def _parse_nfft(self, n_fft=None):
        if n_fft is None:
            return self.n_fft
        return n_fft

    def _parse_nfft_coh(self, n_fft_coh=None):
        if n_fft_coh is None:
            return self.n_fft_coh
        return n_fft_coh

    def _check_ds(self, raw_ds, out_ds):
        """
        Check that the attributes between two datasets match up.

        Parameters
        ----------
        raw_ds : xarray.Dataset
          Input dataset
        out_ds : xarray.Dataset
          Dataset to append `raw_ds` to. If None is supplied, this
          dataset is created from `raw_ds`.

        Returns
        -------
        out_ds : xarray.Dataset
        """

        for v in raw_ds.data_vars:
            if np.any(np.array(raw_ds[v].shape) == 0):
                raise RuntimeError(f"{v} cannot be averaged "
                                   "because it is empty.")
        if 'DutyCycle_NBurst' in raw_ds.attrs and \
                raw_ds.attrs['DutyCycle_NBurst'] < self.n_bin:
            warnings.warn(f"The averaging interval (n_bin = {self.n_bin})"
                          "is larger than the burst interval "
                          "(NBurst = {dat.attrs['DutyCycle_NBurst']})")
        if raw_ds.fs != self.fs:
            raise Exception(f"The input data sample rate ({raw_ds.fs}) does not "
                            "match the sample rate of this binning-object "
                            "({self.fs})")

        if out_ds is None:
            out_ds = type(raw_ds)()

        o_attrs = out_ds.attrs

        props = {}
        props['fs'] = self.fs
        props['n_bin'] = self.n_bin
        props['n_fft'] = self.n_fft
        props['description'] = 'Binned averages calculated from ' \
            'ensembles of size "n_bin"'
        props.update(raw_ds.attrs)

        for ky in props:
            if ky in o_attrs and o_attrs[ky] != props[ky]:
                # The values in out_ds must match `props` (raw_ds.attrs,
                # plus those defined above)
                raise AttributeError(
                    "The attribute '{}' of `out_ds` is inconsistent "
                    "with this `VelBinner` or the input data (`raw_ds`)".format(ky))
            else:
                o_attrs[ky] = props[ky]
        return out_ds

    def _new_coords(self, array):
        """Function for setting up a new xarray.DataArray regardless of how 
        many dimensions the input data-array has
        """
        dims = array.dims
        dims_list = []
        coords_dict = {}
        if len(array.shape) == 1 & ('dir' in array.coords):
            array = array.drop_vars('dir')
        for ky in dims:
            dims_list.append(ky)
            if 'time' in ky:
                coords_dict[ky] = self.mean(array.time.values)
            else:
                coords_dict[ky] = array.coords[ky].values

        return dims_list, coords_dict

    def reshape(self, arr, n_pad=0, n_bin=None):
        """
        Reshape the array `arr` to shape (...,n,n_bin+n_pad).

        Parameters
        ----------
        arr : numpy.ndarray
        n_pad : int
          Is used to add `n_pad`/2 points from the end of the previous
          ensemble to the top of the current, and `n_pad`/2 points
          from the top of the next ensemble to the bottom of the
          current.  Zeros are padded in the upper-left and lower-right
          corners of the matrix (beginning/end of timeseries).  In
          this case, the array shape will be (...,`n`,`n_pad`+`n_bin`)
        n_bin : int (default is self.n_bin)
          Override this binner's n_bin.

        Returns
        -------
        out : numpy.ndarray

        Notes
        -----
        `n_bin` can be non-integer, in which case the output array
        size will be `n_pad`+`n_bin`, and the decimal will
        cause skipping of some data points in `arr`.  In particular,
        every mod(`n_bin`,1) bins will have a skipped point. For
        example:
        - for n_bin=2048.2 every 1/5 bins will have a skipped point.
        - for n_bin=4096.9 every 9/10 bins will have a skipped point.
        """

        n_bin = self._parse_nbin(n_bin)
        npd0 = int(n_pad // 2)
        npd1 = int((n_pad + 1) // 2)
        shp = self._outshape(arr.shape, n_pad=0, n_bin=n_bin)
        out = np.zeros(
            self._outshape(arr.shape, n_pad=n_pad, n_bin=n_bin),
            dtype=arr.dtype)
        if np.mod(n_bin, 1) == 0:
            # n_bin needs to be int
            n_bin = int(n_bin)
            # If n_bin is an integer, we can do this simply.
            out[..., npd0: n_bin + npd0] = (
                arr[..., :(shp[-2] * shp[-1])]).reshape(shp, order='C')
        else:
            inds = (np.arange(np.prod(shp[-2:])) * n_bin // int(n_bin)
                    ).astype(int)
            n_bin = int(n_bin)
            out[..., npd0:n_bin + npd0] = (arr[..., inds]
                                           ).reshape(shp, order='C')
            n_bin = int(n_bin)
        if n_pad != 0:
            out[..., 1:, :npd0] = out[..., :-1, n_bin:n_bin + npd0]
            out[..., :-1, -npd1:] = out[..., 1:, npd0:npd0 + npd1]

        return out

    def detrend(self, arr, axis=-1, n_pad=0, n_bin=None):
        """
        Reshape the array `arr` to shape (...,n,n_bin+n_pad)
        and remove the best-fit trend line from each bin.

        Parameters
        ----------
        arr : numpy.ndarray
        axis : int (default is -1)
          Axis along which to take mean
        n_pad : int (default is 0)
          Is used to add `n_pad`/2 points from the end of the previous
          ensemble to the top of the current, and `n_pad`/2 points
          from the top of the next ensemble to the bottom of the
          current.  Zeros are padded in the upper-left and lower-right
          corners of the matrix (beginning/end of timeseries).  In
          this case, the array shape will be (...,`n`,`n_pad`+`n_bin`)
        n_bin : int (default is self.n_bin)
          Override this binner's n_bin.

        Returns
        -------
        out : numpy.ndarray
        """

        return detrend(self.reshape(arr, n_pad=n_pad, n_bin=n_bin), axis=axis)

    def demean(self, arr, axis=-1, n_pad=0, n_bin=None):
        """
        Reshape the array `arr` to shape (...,n,n_bin+n_pad)
        and remove the mean from each bin.

        Parameters
        ----------
        arr : numpy.ndarray
        axis : int (default is -1)
          Axis along which to take mean
        n_pad : int (default is 0)
          Is used to add `n_pad`/2 points from the end of the previous
          ensemble to the top of the current, and `n_pad`/2 points
          from the top of the next ensemble to the bottom of the
          current.  Zeros are padded in the upper-left and lower-right
          corners of the matrix (beginning/end of timeseries).  In
          this case, the array shape will be (...,`n`,`n_pad`+`n_bin`)
        n_bin : int (default is self.n_bin)
          Override this binner's n_bin.

        Returns
        -------
        out : numpy.ndarray
        """

        dt = self.reshape(arr, n_pad=n_pad, n_bin=n_bin)
        return dt - np.nanmean(dt, axis)[..., None]

    def mean(self, arr, axis=-1, n_bin=None):
        """
        Reshape the array `arr` to shape (...,n,n_bin+n_pad)
        and take the mean of each bin along the specified `axis`.

        Parameters
        ----------
        arr : numpy.ndarray
        axis : int (default is -1)
          Axis along which to take mean
        n_bin : int (default is self.n_bin)
          Override this binner's n_bin.

        Returns
        -------
        out : numpy.ndarray
        """

        if np.issubdtype(arr.dtype, np.datetime64):
            return epoch2dt64(self.mean(dt642epoch(arr), axis=axis, n_bin=n_bin))
        if axis != -1:
            arr = np.swapaxes(arr, axis, -1)
        n_bin = self._parse_nbin(n_bin)
        tmp = self.reshape(arr, n_bin=n_bin)

        return np.nanmean(tmp, -1)

    def var(self, arr, axis=-1, n_bin=None):
        """
        Reshape the array `arr` to shape (...,n,n_bin+n_pad)
        and take the variance of each bin along the specified `axis`.

        Parameters
        ----------
        arr : numpy.ndarray
        axis : int (default is -1)
          Axis along which to take variance
        n_bin : int (default is self.n_bin)
          Override this binner's n_bin.

        Returns
        -------
        out : numpy.ndarray
        """

        return self.reshape(arr, n_bin=n_bin).var(axis)

    def std(self, arr, axis=-1, n_bin=None):
        """
        Reshape the array `arr` to shape (...,n,n_bin+n_pad)
        and take the standard deviation of each bin along the 
        specified `axis`.

        Parameters
        ----------
        arr : numpy.ndarray
        axis : int (default is -1)
          Axis along which to take std dev
        n_bin : int (default is self.n_bin)
          Override this binner's n_bin.

        Returns
        -------
        out : numpy.ndarray
        """

        return self.reshape(arr, n_bin=n_bin).std(axis)

    def calc_psd_base(self, dat, fs=None, window='hann', noise=0,
                      n_bin=None, n_fft=None, n_pad=None, step=None):
        """
        Calculate power spectral density of `dat`

        Parameters
        ----------
        dat : xarray.DataArray
          The raw dataArray of which to calculate the psd.
        fs : float (optional)
          The sample rate (Hz).
        window : str
          String indicating the window function to use (default: 'hanning').
        noise  : float
          The white-noise level of the measurement (in the same units
          as `dat`).
        n_bin : int
          n_bin of veldat2, number of elements per bin if 'None' is taken 
          from VelBinner
        n_fft : int
          n_fft of veldat2, number of elements per bin if 'None' is taken 
          from VelBinner
        n_pad : int (optional)
          The number of values to pad with zero (default: 0)
        step : int (optional)
          Controls amount of overlap in fft (default: the step size is
          chosen to maximize data use, minimize nens, and have a
          minimum of 50% overlap.).

        Returns
        -------
        out : numpy.ndarray
          The power spectral density of `dat`

        Notes
        -----
        PSD's are calculated based on sample rate units
        """

        fs = self._parse_fs(fs)
        n_bin = self._parse_nbin(n_bin)
        n_fft = self._parse_nfft(n_fft)
        if n_pad is None:
            n_pad = min(n_bin - n_fft, n_fft)
        out = np.empty(self._outshape_fft(dat.shape, n_fft=n_fft, n_bin=n_bin))
        # The data is detrended in psd, so we don't need to do it here.
        dat = self.reshape(dat, n_pad=n_pad)

        for slc in slice1d_along_axis(dat.shape, -1):
            out[slc] = psd(dat[slc], n_fft, fs,
                           window=window, step=step)
        if noise != 0:
            out -= noise**2 / (fs/2)
            # Make sure all values of the PSD are >0 (but still small):
            out[out < 0] = np.min(np.abs(out)) / 100
        return out

    def calc_csd_base(self, dat1, dat2, fs=None, window='hann',
                      n_fft=None, n_bin=None):
        """
        Calculate the cross power spectral density of `dat`.

        Parameters
        ----------
        dat1 : numpy.ndarray
          The first (shorter, if applicable) raw dataArray of which to 
          calculate the cpsd.
        dat2 : numpy.ndarray
          The second (the shorter, if applicable) raw dataArray of which to 
          calculate the cpsd.
        fs : float (optional)
          The sample rate (rad/s or Hz).
        window : str
          String indicating the window function to use (default: 'hanning').
        n_fft : int
          n_fft of veldat2, number of elements per bin if 'None' is taken 
          from VelBinner
        n_bin : int
          n_bin of veldat2, number of elements per bin if 'None' is taken 
          from VelBinner

        Returns
        -------
        out : numpy.ndarray
          The cross power spectral density of `dat1` and `dat2`

        Notes
        -----
        PSD's are calculated based on sample rate units

        The two velocity inputs do not have to be perfectly synchronized, but 
        they should have the same start and end timestamps
        """

        fs = self._parse_fs(fs)
        if n_fft is None:
            n_fft = self.n_fft_coh
        # want each slice to carry the same timespan
        n_bin2 = self._parse_nbin(n_bin)  # bins for shorter array
        n_bin1 = int(dat1.shape[-1]/(dat2.shape[-1]/n_bin2))

        oshp = self._outshape_fft(dat1.shape, n_fft=n_fft, n_bin=n_bin1)
        oshp[-2] = np.min([oshp[-2], int(dat2.shape[-1] // n_bin2)])

        # The data is detrended in psd, so we don't need to do it here:
        dat1 = self.reshape(dat1, n_pad=n_fft)
        dat2 = self.reshape(dat2, n_pad=n_fft)
        out = np.empty(oshp, dtype='c{}'.format(dat1.dtype.itemsize * 2))
        if dat1.shape == dat2.shape:
            cross = cpsd
        else:
            cross = cpsd_quasisync
        for slc in slice1d_along_axis(out.shape, -1):
            out[slc] = cross(dat1[slc], dat2[slc], n_fft,
                             fs, window=window)
        return out

    def calc_freq(self, fs=None, units='rad/s', n_fft=None, coh=False):
        """
        Calculate the ordinary or radial frequency vector for the PSDs

        Parameters
        ----------
        fs : float (optional)
          The sample rate (Hz).
        units : string
          Frequency units in either Hz or rad/s (f or omega)
        coh : bool
          Calculate the frequency vector for coherence/cross-spectra
          (default: False) i.e. use self.n_fft_coh instead of
          self.n_fft.
        n_fft : int
          n_fft of veldat2, number of elements per bin if 'None' is taken 
          from VelBinner

        Returns
        -------
        out: numpy.ndarray
          Spectrum frequency array in units of 'Hz' or 'rad/s'
        """

        if n_fft is None:
            n_fft = self.n_fft
            if coh:
                n_fft = self.n_fft_coh

        fs = self._parse_fs(fs)

        if ('Hz' not in units) and ('rad' not in units):
            raise Exception('Valid fft frequency vector units are Hz \
                            or rad/s')

        if 'rad' in units:
            return psd_freq(n_fft, 2*np.pi*fs)
        else:
            return psd_freq(n_fft, fs)



from .vector import _earth2principal, _euler2orient as euler2orient
from .base import _beam2inst
from . import base as rotb
import numpy as np
import warnings
from numpy.linalg import inv


def _inst2earth(adcpo, reverse=False, rotate_vars=None, force=False):
    """
    Rotate data in an ADCP object to the earth from the instrument
    frame (or vice-versa).

    Parameters
    ----------
    adcpo : The adv object containing the data.

    reverse : bool (default: False)
           If True, this function performs the inverse rotation
           (earth->inst).

    rotate_vars : iterable
      The list of variables to rotate. By default this is taken from
      adcpo.props['rotate_vars'].

    force : Do not check which frame the data is in prior to
      performing this rotation.

    """

    if reverse:
        # The transpose of the rotation matrix gives the inverse
        # rotation, so we simply reverse the order of the einsum:
        sumstr = 'jik,j...k->i...k'
        cs_now = 'earth'
        cs_new = 'inst'
    else:
        sumstr = 'ijk,j...k->i...k'
        cs_now = 'inst'
        cs_new = 'earth'

    # if ADCP is upside down
    if adcpo.orientation == 'down':
        down = True
    else:  # orientation = 'up' or 'AHRS'
        down = False

    if rotate_vars is None:
        if 'rotate_vars' in adcpo.attrs:
            rotate_vars = adcpo.rotate_vars
        else:
            rotate_vars = ['vel']

    cs = adcpo.coord_sys.lower()
    if not force:
        if cs == cs_new:
            print("Data is already in the '%s' coordinate system" % cs_new)
            return
        elif cs != cs_now:
            raise ValueError(
                "Data must be in the '%s' frame when using this function" %
                cs_now)

    if 'orientmat' in adcpo:
        rmat = adcpo['orientmat'].values
    else:
        rmat = euler2orient(adcpo['heading'].values, adcpo['pitch'].values,
                            adcpo['roll'].values)

    # Take the transpose of the orientation to get the inst->earth rotation
    # matrix.
    rmat = np.rollaxis(rmat, 1)

    _dcheck = rotb._check_rotmat_det(rmat)
    if not _dcheck.all():
        warnings.warn("Invalid orientation matrix (determinant != 1) at indices: {}. "
                      "If rotated, data at these indices will be erroneous."
                      .format(np.nonzero(~_dcheck)[0]), UserWarning)

    # The dictionary of rotation matrices for different sized arrays.
    rmd = {3: rmat, }

    # The 4-row rotation matrix assume that rows 0,1 are u,v,
    # and 2,3 are independent estimates of w.
    tmp = rmd[4] = np.zeros((4, 4, rmat.shape[-1]), dtype=np.float64)
    tmp[:3, :3] = rmat
    # Copy row 2 to 3
    tmp[3, :2] = rmat[2, :2]
    tmp[3, 3] = rmat[2, 2]
    # Extend rows 0,1
    tmp[0, 2:] = rmat[0, 2] / 2
    tmp[1, 2:] = rmat[1, 2] / 2

    if reverse:
        # 3-element inverse handled by sumstr definition (transpose)
        rmd[4] = np.moveaxis(inv(np.moveaxis(rmd[4], -1, 0)), 0, -1)

    for nm in rotate_vars:
        dat = adcpo[nm].values
        n = dat.shape[0]
        # Nortek documents sign change for upside-down instruments
        if down:
            # This is equivalent to adding 180 degrees to roll axis in _calc_omat()
            sign = np.array([1, -1, -1, -1], ndmin=dat.ndim).T
            signIMU = np.array([1, -1, -1], ndmin=dat.ndim).T
            if not reverse:
                if n == 3:
                    dat = np.einsum(sumstr, rmd[3], signIMU*dat)
                elif n == 4:
                    dat = np.einsum('ijk,j...k->i...k', rmd[4], sign*dat)
                else:
                    raise Exception("The entry {} is not a vector, it cannot"
                                    "be rotated.".format(nm))

            elif reverse:
                if n == 3:
                    dat = signIMU*np.einsum(sumstr, rmd[3], dat)
                elif n == 4:
                    dat = sign*np.einsum('ijk,j...k->i...k', rmd[4], dat)
                else:
                    raise Exception("The entry {} is not a vector, it cannot"
                                    "be rotated.".format(nm))

        else:  # 'up' and AHRS
            if n == 3:
                dat = np.einsum(sumstr, rmd[3], dat)
            elif n == 4:
                dat = np.einsum('ijk,j...k->i...k', rmd[4], dat)
            else:
                raise Exception("The entry {} is not a vector, it cannot"
                                "be rotated.".format(nm))
        adcpo[nm].values = dat.copy()

    adcpo = rotb._set_coords(adcpo, cs_new)

    return adcpo



"""Module containing functions to clean data
"""
import numpy as np
import xarray as xr
from scipy.signal import medfilt
from ..tools.misc import medfiltnan
from ..rotate.api import rotate2
from ..rotate.base import _make_model, quaternion2orient


def set_range_offset(ds, h_deploy):
    """
    Adds an instrument's height above seafloor (for an up-facing instrument)
    or depth below water surface (for a down-facing instrument) to the range
    coordinate. Also adds an attribute to the Dataset with the current
    "h_deploy" distance.

    Parameters
    ----------
    ds : xarray.Dataset
      The adcp dataset to ajust 'range' on
    h_deploy : numeric
      Deployment location in the water column, in [m]

    Returns
    -------
    None, operates "in place"

    Notes
    -----
    `Center of bin 1 = h_deploy + blank_dist + cell_size`

    Nortek doesn't take `h_deploy` into account, so the range that DOLfYN
    calculates distance is from the ADCP transducers. TRDI asks for `h_deploy`
    input in their deployment software and is thereby known by DOLfYN.

    If the ADCP is mounted on a tripod on the seafloor, `h_deploy` will be
    the height of the tripod +/- any extra distance to the transducer faces.
    If the instrument is vessel-mounted, `h_deploy` is the distance between
    the surface and downward-facing ADCP's transducers.
    """

    r = [s for s in ds.dims if 'range' in s]
    for val in r:
        ds[val] = ds[val].values + h_deploy
        ds[val].attrs['units'] = 'm'

    if hasattr(ds, 'h_deploy'):
        ds.attrs['h_deploy'] += h_deploy
    else:
        ds.attrs['h_deploy'] = h_deploy


def find_surface(ds, thresh=10, nfilt=None):
    """
    Find the surface (water level or seafloor) from amplitude data and
    adds the variable "depth" to the input Dataset.

    Parameters
    ----------
    ds : xarray.Dataset
      The full adcp dataset
    thresh : int
      Specifies the threshold used in detecting the surface.
      (The amount that amplitude must increase by near the surface for it to
      be considered a surface hit)
    nfilt : int
      Specifies the width of the median filter applied, must be odd

    Returns
    -------
    None, operates "in place"
    """

    # This finds the maximum of the echo profile:
    inds = np.argmax(ds.amp.values, axis=1)
    # This finds the first point that increases (away from the profiler) in
    # the echo profile
    edf = np.diff(ds.amp.values.astype(np.int16), axis=1)
    inds2 = np.max((edf < 0) *
                   np.arange(ds.vel.shape[1] - 1,
                             dtype=np.uint8)[None, :, None], axis=1) + 1

    # Calculate the depth of these quantities
    d1 = ds.range.values[inds]
    d2 = ds.range.values[inds2]
    # Combine them:
    D = np.vstack((d1, d2))
    # Take the median value as the estimate of the surface:
    d = np.median(D, axis=0)

    # Throw out values that do not increase near the surface by *thresh*
    for ip in range(ds.vel.shape[1]):
        itmp = np.min(inds[:, ip])
        if (edf[itmp:, :, ip] < thresh).all():
            d[ip] = np.NaN

    if nfilt:
        dfilt = medfiltnan(d, nfilt, thresh=.4)
        dfilt[dfilt == 0] = np.NaN
        d = dfilt

    ds['depth'] = xr.DataArray(d, dims=['time'], attrs={
                               'units': 'm'}).astype('float32')


def find_surface_from_P(ds, salinity=35):
    """
    Calculates the distance to the water surface. Temperature and salinity
    are used to calculate seawater density, which is in turn used with the
    pressure data to calculate depth.

    Parameters
    ----------
    ds : xarray.Dataset
      The full adcp dataset
    salinity: numeric
      Water salinity in psu

    Returns
    -------
    None, operates "in place" and adds the variables "water_density" and
    "depth" to the input dataset.

    Notes
    -----
    Requires that the instrument's pressure sensor was calibrated/zeroed
    before deployment to remove atmospheric pressure.

    Calculates seawater density at normal atmospheric pressure according
    to the UNESCO 1981 equation of state. Does not include hydrostatic pressure.
    """

    # Density calcation
    T = ds.temp.values
    S = salinity
    # standard mean ocean water
    rho_smow = 999.842594 + 6.793953e-2*T - 9.095290e-3*T**2 + \
        1.001685e-4*T**3 - 1.120083e-6*T**4 + 6.536332e-9*T**5
    # at water surface
    B1 = 8.2449e-1 - 4.0899e-3*T + 7.6438e-5*T**2 - 8.2467e-7*T**3 + 5.3875e-9*T**4
    C1 = -5.7246e-3 + 1.0227e-4*T - 1.6546e-6*T**2
    d0 = 4.8314e-4
    rho_atm0 = rho_smow + B1*S + C1*S**1.5 + d0*S**2

    # Depth = pressure (conversion from dbar to MPa) / water weight
    d = (ds.pressure*10000)/(9.81*rho_atm0)

    if hasattr(ds, 'h_deploy'):
        d += ds.h_deploy
        description = "Water depth to seafloor"
    else:
        description = "Water depth to ADCP"

    ds['water_density'] = xr.DataArray(
        rho_atm0.astype('float32'),
        dims=['time'],
        attrs={'units': 'kg/m^3',
               'description': 'Water density according to UNESCO 1981 equation of state'})
    ds['depth'] = xr.DataArray(d.astype('float32'), dims=['time'], attrs={
                               'units': 'm', 'description': description})


def nan_beyond_surface(ds, val=np.nan, inplace=False):
    """
    Mask the values of 3D data (vel, amp, corr, echo) that are beyond the surface.

    Parameters
    ----------
    ds : xarray.Dataset
      The adcp dataset to clean
    val : nan or numeric
      Specifies the value to set the bad values to (default np.nan).
    inplace : bool (default: False)
      When True the existing data object is modified. When False
      a copy is returned.

    Returns
    -------
    ds : xarray.Dataset
      Sets the adcp dataset where relevant arrays with values greater than `depth`
      set to NaN

    Notes
    -----
    Surface interference expected to happen at 
    `distance > range * cos(beam angle) - cell size`
    """

    if not inplace:
        ds = ds.copy(deep=True)

    # Get all variables with 'range' coordinate
    var = [h for h in ds.keys() if any(s for s in ds[h].dims if 'range' in s)]

    if 'nortek' in _make_model(ds):
        beam_angle = 25 * (np.pi/180)
    else:  # TRDI
        try:
            beam_angle = ds.beam_angle
        except:
            beam_angle = 20 * (np.pi/180)

    # Surface interference distance calculated from distance of transducers to surface
    if hasattr(ds, 'h_deploy'):
        range_limit = ((ds.depth-ds.h_deploy) * np.cos(beam_angle) -
                       ds.cell_size) + ds.h_deploy
    else:
        range_limit = ds.depth * np.cos(beam_angle) - ds.cell_size

    bds = ds.range > range_limit

    # Echosounder data needs only be trimmed at water surface
    if 'echo' in var:
        bds_echo = ds.range_echo > ds.depth
        ds['echo'].values[..., bds_echo] = val
        var.remove('echo')

    # Correct rest of "range" data for surface interference
    for nm in var:
        a = ds[nm].values
        try:  # float dtype
            a[..., bds] = val
        except:  # int dtype
            a[..., bds] = 0
        ds[nm].values = a

    if not inplace:
        return ds


def correlation_filter(ds, thresh=50, inplace=False):
    """
    Filters out data where correlation is below a threshold in the 
    along-beam correlation data.

    Parameters
    ----------
    ds : xarray.Dataset
      The adcp dataset to clean.
    thresh : numeric
      The maximum value of correlation to screen, in counts or %
    inplace : bool (default: False)
      When True the existing data object is modified. When False
      a copy is returned.

    Returns
    -------
    ds : xarray.Dataset
      Elements in velocity, correlation, and amplitude are removed if below the 
      correlation threshold

    Notes
    -----
    Does not edit correlation or amplitude data.
    """

    if not inplace:
        ds = ds.copy(deep=True)

    # 4 or 5 beam
    if hasattr(ds, 'vel_b5'):
        tag = ['', '_b5']
    else:
        tag = ['']

    # copy original ref frame
    coord_sys_orig = ds.coord_sys

    # correlation is always in beam coordinates
    rotate2(ds, 'beam', inplace=True)
    # correlation is always in beam coordinates
    for tg in tag:
        mask = ds['corr'+tg].values <= thresh

        for var in ['vel', 'corr', 'amp']:
            try:
                ds[var+tg].values[mask] = np.nan
            except:
                ds[var+tg].values[mask] = 0
            ds[var+tg].attrs['Comments'] = 'Filtered of data with a correlation value below ' + \
                str(thresh) + ds.corr.units

    rotate2(ds, coord_sys_orig, inplace=True)

    if not inplace:
        return ds


def medfilt_orient(ds, nfilt=7):
    """
    Median filters the orientation data (heading-pitch-roll or quaternions)

    Parameters
    ----------
    ds : xarray.Dataset
      The adcp dataset to clean
    nfilt : numeric
      The length of the median-filtering kernel
      *nfilt* must be odd.

    Return
    ------
    ds : xarray.Dataset
      The adcp dataset with the filtered orientation data

    See Also
    --------
    scipy.signal.medfilt()
    """

    ds = ds.copy(deep=True)

    if getattr(ds, 'has_imu'):
        q_filt = np.zeros(ds.quaternions.shape)
        for i in range(ds.quaternions.q.size):
            q_filt[i] = medfilt(ds.quaternions[i].values, nfilt)
        ds.quaternions.values = q_filt

        ds['orientmat'] = quaternion2orient(ds.quaternions)
        return ds

    else:
        # non Nortek AHRS-equipped instruments
        do_these = ['pitch', 'roll', 'heading']
        for nm in do_these:
            ds[nm].values = medfilt(ds[nm].values, nfilt)

        return ds.drop_vars('orientmat')


def val_exceeds_thresh(var, thresh=5, val=np.nan):
    """
    Find values of a variable that exceed a threshold value,
    and assign "val" to the velocity data where the threshold is
    exceeded.

    Parameters
    ----------
    var : xarray.DataArray
      Variable to clean
    thresh : numeric
      The maximum value of velocity to screen
    val : nan or numeric
      Specifies the value to set the bad values to (default np.nan)

    Returns
    -------
    ds : xarray.Dataset
      The adcp dataset with datapoints beyond thresh are set to `val`

    """
    var = var.copy(deep=True)

    bd = np.zeros(var.shape, dtype='bool')
    bd |= (np.abs(var.values) > thresh)

    var.values[bd] = val

    return var


def fillgaps_time(var, method='cubic', maxgap=None):
    """
    Fill gaps (nan values) in var across time using the specified method

    Parameters
    ----------
    var : xarray.DataArray
      The variable to clean
    method : string
      Interpolation method to use
    maxgap : numeric
      Maximum gap of missing data to interpolate across

    Returns
    -------
    out : xarray.DataArray
      The input DataArray 'var' with gaps in 'var' interpolated across time

    See Also
    --------
    xarray.DataArray.interpolate_na()
    """

    time_dim = [t for t in var.dims if 'time' in t][0]

    return var.interpolate_na(dim=time_dim, method=method,
                              use_coordinate=True,
                              limit=maxgap)


def fillgaps_depth(var, method='cubic', maxgap=None):
    """
    Fill gaps (nan values) in var along the depth profile using the specified method

    Parameters
    ----------
    var : xarray.DataArray
      The variable to clean
    method : string
      Interpolation method to use
    maxgap : int
      Maximum gap of missing data to interpolate across

    Returns
    -------
    out : xarray.DataArray
      The input DataArray 'var' with gaps in 'var' interpolated across depth

    See Also
    --------
    xarray.DataArray.interpolate_na()
    """

    range_dim = [t for t in var.dims if 'range' in t][0]

    return var.interpolate_na(dim=range_dim, method=method,
                              use_coordinate=False,
                              limit=maxgap)



import numpy as np
from ..velocity import VelBinner
import warnings
from ..tools.misc import slice1d_along_axis, _nans_like
from scipy.special import cbrt
import xarray as xr


class ADVBinner(VelBinner):
    """
    A class that builds upon `VelBinner` for calculating turbulence 
    statistics and velocity spectra from ADV data

    Parameters
    ----------
    n_bin : int
      The length of each `bin`, in number of points, for this averaging
      operator.
    fs : int
      Instrument sampling frequency in Hz
    n_fft : int (optional, default: n_fft = n_bin)
      The length of the FFT for computing spectra (must be <= n_bin)
    n_fft_coh : int
      Number of data points to use for coherence and cross-spectra ffts
      Default: `n_fft_coh`=`n_fft`
    noise : float, list or numpy.ndarray
      Instrument's doppler noise in same units as velocity
    """

    def __call__(self, ds, freq_units='rad/s', window='hann'):
        out = type(ds)()
        out = self.do_avg(ds, out)

        noise = ds.get('doppler_noise', [0, 0, 0])
        out['tke_vec'] = self.calc_tke(ds['vel'], noise=noise)
        out['stress_vec'] = self.calc_stress(ds['vel'])

        out['psd'] = self.calc_psd(ds['vel'],
                                   window=window,
                                   freq_units=freq_units,
                                   noise=noise)
        for key in list(ds.attrs.keys()):
            if 'config' in key:
                ds.attrs.pop(key)
        out.attrs = ds.attrs
        out.attrs['n_bin'] = self.n_bin
        out.attrs['n_fft'] = self.n_fft
        out.attrs['n_fft_coh'] = self.n_fft_coh

        return out

    def calc_stress(self, veldat, detrend=True):
        """
        Calculate the stresses (covariances of u,v,w)

        Parameters
        ----------
        veldat : xr.DataArray
            Velocity data array from ADV data. The last dimension is 
            assumed to be time.
        detrend : bool (default: True)
            detrend the velocity data (True), or simply de-mean it
            (False), prior to computing stress. Note: the psd routines
            use detrend, so if you want to have the same amount of
            variance here as there use ``detrend=True``.

        Returns
        -------
        out : xarray.DataArray
        """

        time = self.mean(veldat.time.values)
        vel = veldat.values

        out = np.empty(self._outshape(vel.shape)[:-1],
                       dtype=np.float32)

        if detrend:
            vel = self.detrend(vel)
        else:
            vel = self.demean(vel)

        for idx, p in enumerate(self._cross_pairs):
            out[idx] = np.nanmean(vel[p[0]] * vel[p[1]],
                                  -1, dtype=np.float64
                                  ).astype(np.float32)

        da = xr.DataArray(out.astype('float32'), 
                          dims=veldat.dims, 
                          attrs={'units': 'm^2/^2'})
        da = da.rename({'dir': 'tau'})
        da = da.assign_coords({'tau': ["upvp_", "upwp_", "vpwp_"],
                               'time': time})
        return da

    def calc_csd(self, veldat,
                 freq_units='rad/s',
                 fs=None,
                 window='hann',
                 n_bin=None,
                 n_fft_coh=None):
        """
        Calculate the cross-spectral density of velocity components.

        Parameters
        ----------
        veldat   : xarray.DataArray
          The raw 3D velocity data.
        freq_units : string
          Frequency units of the returned spectra in either Hz or rad/s 
          (`f` or :math:`\\omega`)
        fs : float (optional)
          The sample rate (default: from the binner).
        window : string or array
          Specify the window function.
        n_bin : int (optional)
          The bin-size (default: from the binner).
        n_fft_coh : int (optional)
          The fft size (default: n_fft_coh from the binner).

        Returns
        -------
        csd : xarray.DataArray (3, M, N_FFT)
          The first-dimension of the cross-spectrum is the three
          different cross-spectra: 'uv', 'uw', 'vw'.
        """

        fs = self._parse_fs(fs)
        n_fft = self._parse_nfft_coh(n_fft_coh)
        time = self.mean(veldat.time.values)
        veldat = veldat.values

        out = np.empty(self._outshape_fft(veldat[:3].shape, n_fft=n_fft),
                       dtype='complex')

        # Create frequency vector, also checks whether using f or omega
        coh_freq = self.calc_freq(units=freq_units, coh=True)
        if 'rad' in freq_units:
            fs = 2*np.pi*fs
            freq_units = 'rad/s'
            units = 'm^2/s/rad'
        else:
            freq_units = 'Hz'
            units = 'm^2/s^2/Hz'

        for ip, ipair in enumerate(self._cross_pairs):
            out[ip] = self.calc_csd_base(veldat[ipair[0]],
                                         veldat[ipair[1]],
                                         n_bin=n_bin,
                                         n_fft=n_fft,
                                         window=window)

        csd = xr.DataArray(out,
                           coords={'C': ['Cxy', 'Cxz', 'Cyz'],
                                   'time': time,
                                   'freq': coh_freq},
                           dims=['C', 'time', 'freq'],
                           attrs={'units': units, 'n_fft_coh': n_fft})
        csd['freq'].attrs['units'] = freq_units

        return csd

    def calc_epsilon_LT83(self, psd, U_mag, freq_range=[6.28, 12.57]):
        """
        Calculate the dissipation rate from the PSD

        Parameters
        ----------
        psd : xarray.DataArray (...,time,f)
          The power spectral density
        U_mag : xarray.DataArray (...,time)
          The bin-averaged horizontal velocity [m/s] (from dataset shortcut)
        freq_range : iterable(2)
          The range over which to integrate/average the spectrum, in units 
          of the psd frequency vector (Hz or rad/s)

        Returns
        -------
        epsilon : xarray.DataArray (...,n_time)
          dataArray of the dissipation rate
        
        Notes
        -----
        This uses the `standard` formula for dissipation:

        .. math:: S(k) = \\alpha \\epsilon^{2/3} k^{-5/3} + N

        where :math:`\\alpha = 0.5` (1.5 for all three velocity
        components), `k` is wavenumber, `S(k)` is the turbulent
        kinetic energy spectrum, and `N' is the doppler noise level
        associated with the TKE spectrum.

        With :math:`k \\rightarrow \\omega / U`, then -- to preserve variance --
        :math:`S(k) = U S(\\omega)`, and so this becomes:

        .. math:: S(\\omega) = \\alpha \\epsilon^{2/3} \\omega^{-5/3} U^{2/3} + N
        
        With :math:`k \\rightarrow (2\\pi f) / U`, then

        .. math:: S(\\omega) = \\alpha \\epsilon^{2/3} f^{-5/3} (U/(2\\pi))^{2/3} + N
        
        LT83 : Lumley and Terray, "Kinematics of turbulence convected
        by a random wave field". JPO, 1983, vol13, pp2000-2007.
        """

        freq = psd.freq

        idx = np.where((freq_range[0] < freq) & (freq < freq_range[1]))
        idx = idx[0]

        if freq.units == 'Hz':
            U = U_mag/(2*np.pi)
        else:
            U = U_mag

        a = 0.5
        out = (psd.isel(freq=idx) *
               freq.isel(freq=idx)**(5/3) / a).mean(axis=-1)**(3/2) / U

        out = xr.DataArray(out, name='dissipation_rate',
                           attrs={'units': 'm^2/s^3',
                                  'method': 'LT83'})
        return out

    def calc_epsilon_SF(self, vel_raw, U_mag, fs=None, freq_range=[2., 4.]):
        """
        Calculate dissipation rate using the "structure function" (SF) method
        
        Parameters
        ----------
        vel_raw : xarray.DataArray
          The raw velocity data (1D dimension time) upon 
          which to perform the SF technique. 
        U_mag : xarray.DataArray
          The bin-averaged horizontal velocity (from dataset shortcut)
        fs : float
          The sample rate of `vel_raw` [Hz]
        freq_range : iterable(2)
          The frequency range over which to compute the SF [Hz]
          (i.e. the frequency range within which the isotropic 
          turbulence cascade falls)

        Returns
        -------
        epsilon : xarray.DataArray
          dataArray of the dissipation rate
        """

        veldat = vel_raw.values
        if len(veldat.shape) > 1:
            raise Exception("Function input should be a 1D velocity vector")

        fs = self._parse_fs(fs)
        if freq_range[1] > fs:
            warnings.warn('Max freq_range cannot be greater than fs')

        dt = self.reshape(veldat)
        out = np.empty(dt.shape[:-1], dtype=dt.dtype)
        for slc in slice1d_along_axis(dt.shape, -1):
            up = dt[slc]
            lag = U_mag.values[slc[:-1]] / fs * np.arange(up.shape[0])
            DAA = _nans_like(lag)
            for L in range(int(fs / freq_range[1]), int(fs / freq_range[0])):
                DAA[L] = np.nanmean((up[L:] - up[:-L]) ** 2, dtype=np.float64)
            cv2 = DAA / (lag ** (2 / 3))
            cv2m = np.median(cv2[np.logical_not(np.isnan(cv2))])
            out[slc[:-1]] = (cv2m / 2.1) ** (3 / 2)

        return xr.DataArray(out, name='dissipation_rate',
                            coords=U_mag.coords,
                            dims=U_mag.dims,
                            attrs={'units': 'm^2/s^3',
                                   'method': 'structure function'})

    def _up_angle(self, U_complex):
        """
        Calculate the angle of the turbulence fluctuations.

        Parameters
        ----------
        U_complex  : |np.ndarray| (..., n_time * n_bin)
          The complex, raw horizontal velocity (non-binned)

        Returns
        -------
        theta : |np.ndarray| (..., n_time)
          The angle of the turbulence [rad]
        """

        dt = self.demean(U_complex)
        fx = dt.imag <= 0
        dt[fx] = dt[fx] * np.exp(1j * np.pi)

        return np.angle(np.mean(dt, -1, dtype=np.complex128))

    def _calc_epsTE01_int(self, I_tke, theta):
        """
        The integral, equation A13, in [TE01].

        Parameters
        ----------
        I_tke : |np.ndarray|
          (beta in TE01) is the turbulence intensity ratio:
          \\sigma_u / V
        theta : |np.ndarray|
          is the angle between the mean flow and the primary axis of
          velocity fluctuations
        """

        x = np.arange(-20, 20, 1e-2)  # I think this is a long enough range.
        out = np.empty_like(I_tke.flatten())
        for i, (b, t) in enumerate(zip(I_tke.flatten(), theta.flatten())):
            out[i] = np.trapz(
                cbrt(x**2 - 2/b*np.cos(t)*x + b**(-2)) *
                np.exp(-0.5 * x ** 2), x)

        return out.reshape(I_tke.shape) * \
            (2 * np.pi) ** (-0.5) * I_tke ** (2 / 3)

    def calc_epsilon_TE01(self, dat_raw, dat_avg, freq_range=[6.28, 12.57]):
        """
        Calculate the dissipation rate according to TE01.

        Parameters
        ----------
        dat_raw : xarray.Dataset
          The raw (off the instrument) adv dataset
        dat_avg : xarray.Dataset
          The bin-averaged adv dataset (calc'd from 'calc_turbulence' or
          'do_avg'). The spectra (psd) and basic turbulence statistics 
          ('tke_vec' and 'stress_vec') must already be computed.
        freq_range : iterable(2)
          The range over which to integrate/average the spectrum, in units 
          of the psd frequency vector (Hz or rad/s)

        Notes
        -----
        TE01 : Trowbridge, J and Elgar, S, "Turbulence measurements in
        the Surf Zone". JPO, 2001, vol31, pp2403-2417.
        """

        # Assign local names
        U_mag = dat_avg.velds.U_mag.values
        I_tke = dat_avg.velds.I_tke.values
        theta = np.angle(dat_avg.velds.U.values) - \
            self._up_angle(dat_raw.velds.U.values)
        freq = dat_avg['psd'].freq.values

        # Calculate constants
        alpha = 1.5
        intgrl = self._calc_epsTE01_int(I_tke, theta)

        # Index data to be used
        inds = (freq_range[0] < freq) & (freq < freq_range[1])
        psd = dat_avg.psd[..., inds].values
        freq = freq[inds].reshape([1] * (dat_avg.psd.ndim - 2) + [sum(inds)])

        # Estimate values
        # u & v components (equation 6)
        out = (np.nanmean((psd[0] + psd[1]) * freq**(5/3), -1) /
               (21/55 * alpha * intgrl))**(3/2) / U_mag

        # Add w component
        out += (np.nanmean(psd[2] * freq**(5/3), -1) /
                (12/55 * alpha * intgrl))**(3/2) / U_mag

        # Average the two estimates
        out *= 0.5

        return xr.DataArray(out, name='dissipation_rate',
                            coords={'time': dat_avg.psd.time},
                            dims='time',
                            attrs={'units': 'm^2/s^3',
                                   'method': 'TE01'})

    def calc_L_int(self, a_cov, U_mag, fs=None):
        """
        Calculate integral length scales.

        Parameters
        ----------
        a_cov : xarray.DataArray
          The auto-covariance array (i.e. computed using `calc_acov`).
        U_mag : xarray.DataArray
          The bin-averaged horizontal velocity (from dataset shortcut)
        fs : float
          The raw sample rate

        Returns
        -------
        L_int : |np.ndarray| (..., n_time)
          The integral length scale (T_int*U_mag).

        Notes
        ----
        The integral time scale (T_int) is the lag-time at which the
        auto-covariance falls to 1/e.

        If T_int is not reached, L_int will default to '0'.
        """

        acov = a_cov.values
        fs = self._parse_fs(fs)

        scale = np.argmin((acov/acov[..., :1]) > (1/np.e), axis=-1)
        L_int = U_mag.values / fs * scale

        return xr.DataArray(L_int, name='L_int',
                            coords={'dir': a_cov.dir, 'time': a_cov.time},
                            attrs={'units': 'm'})


def calc_turbulence(ds_raw, n_bin, fs, n_fft=None, freq_units='rad/s', window='hann'):
    """
    Functional version of `ADVBinner` that computes a suite of turbulence 
    statistics for the input dataset, and returns a `binned` data object.
    
    Parameters
    ----------
    ds_raw : xarray.Dataset
      The raw adv datset to `bin`, average and compute
      turbulence statistics of.
    freq_units : string
      Frequency units of the returned spectra in either Hz or rad/s 
      (`f` or :math:`\\omega`)
    window : 1, None, 'hann'
      The window to use for calculating power spectral densities
    
    Returns
    -------
    ds : xarray.Dataset
      Returns an 'binned' (i.e. 'averaged') data object. All
      fields (variables) of the input data object are averaged in n_bin
      chunks. This object also computes the following items over
      those chunks:

      - tke_vec : The energy in each component, each components is
        alternatively accessible as:
        :attr:`upup_ <dolfyn.velocity.Velocity.upup_>`,
        :attr:`vpvp_ <dolfyn.velocity.Velocity.vpvp_>`,
        :attr:`wpwp_ <dolfyn.velocity.Velocity.wpwp_>`)

      - stress_vec : The Reynolds stresses, each component is
        alternatively accessible as:
        :attr:`upwp_ <dolfyn.data.velocity.Velocity.upwp_>`,
        :attr:`vpwp_ <dolfyn.data.velocity.Velocity.vpwp_>`,
        :attr:`upvp_ <dolfyn.data.velocity.Velocity.upvp_>`)

      - U_std : The standard deviation of the horizontal
        velocity `U_mag`.
        
      - psd : DataArray containing the spectra of the velocity
        in radial frequency units. The data-array contains:
        - vel : the velocity spectra array (m^2/s/rad))
        - omega : the radial frequncy (rad/s)
    """
    
    calculator = ADVBinner(n_bin, fs, n_fft=n_fft)

    return calculator(ds_raw, freq_units=freq_units, window=window)
    


import numpy as np
nan = np.nan


class _VarAtts():
    """A data variable attributes class.

    Parameters
    ----------
    dims : (list, optional)
        The dimensions of the array other than the 'time'
        dimension. By default the time dimension is appended to the
        end. To specify a point to place it, place 'n' in that
        location.
    dtype : (type, optional)
        The data type of the array to create (default: float32).
    group : (string, optional)
        The data group to which this variable should be a part
        (default: 'main').
    view_type : (type, optional)
        Specify a numpy view to cast the array into.
    default_val : (numeric, optional)
        The value to initialize with (default: use an empty array).
    offset : (numeric, optional)
        The offset, 'b', by which to adjust the data when converting to
        scientific units.
    factor : (numeric, optional)
        The factor, 'm', by which to adjust the data when converting to
        scientific units.
    title_name : (string, optional)
        The name of the variable.
    units : (string, optional)
        The units of this variable.
    dim_names : (list, optional)
        A list of names for each dimension of the array.

    """

    def __init__(self, dims=[], dtype=None, group='main',
                 view_type=None, default_val=None,
                 offset=0, factor=1,
                 title_name=None, units=None, dim_names=None,
                 ):
        self.dims = list(dims)
        if dtype is None:
            dtype = np.float32
        self.dtype = dtype
        self.group = group
        self.view_type = view_type
        self.default_val = default_val
        self.offset = offset
        self.factor = factor
        self.title_name = title_name
        self.units = units
        self.dim_names = dim_names

    def shape(self, **kwargs):
        a = list(self.dims)
        hit = False
        for ky in kwargs:
            if ky in self.dims:
                hit = True
                a[a.index(ky)] = kwargs[ky]
        if hit:
            return a
        else:
            return self.dims + [kwargs['n']]

    def _empty_array(self, **kwargs):
        out = np.zeros(self.shape(**kwargs), dtype=self.dtype)
        try:
            out[:] = np.NaN
        except:
            pass
        if self.view_type is not None:
            out = out.view(self.view_type)
        return out

    def sci_func(self, data):
        """Scale the data to scientific units.

        Parameters
        ----------
        data : :class:`<numpy.ndarray>`
            The data to scale.

        Returns
        -------
        retval : {None, data}
          If this funciton modifies the data in place it returns None,
          otherwise it returns the new data object.
        """
        if self.offset != 0:
            data += self.offset
        if self.factor != 1:
            data *= self.factor
            return data


vec_data = {
    'AnaIn2LSB': _VarAtts(dims=[],
                          dtype=np.uint8,
                          group='sys',
                          units='',
                          ),
    'Count': _VarAtts(dims=[],
                      dtype=np.uint8,
                      group='sys',
                      units='',
                      ),
    'PressureMSB': _VarAtts(dims=[],
                            dtype=np.uint8,
                            group='data_vars',
                            units='dbar',
                            ),
    'AnaIn2MSB': _VarAtts(dims=[],
                          dtype=np.uint8,
                          group='sys',
                          units='',
                          ),
    'PressureLSW': _VarAtts(dims=[],
                            dtype=np.uint16,
                            group='data_vars',
                            units='dbar',
                            ),
    'AnaIn1': _VarAtts(dims=[],
                       dtype=np.uint16,
                       group='sys',
                       units=''
                       ),
    'vel': _VarAtts(dims=[3],
                    dtype=np.float32,
                    group='data_vars',
                    factor=0.001,
                    default_val=nan,
                    units='m/s',
                    ),
    'amp': _VarAtts(dims=[3],
                    dtype=np.uint8,
                    group='data_vars',
                    units='dB',
                    ),
    'corr': _VarAtts(dims=[3],
                     dtype=np.uint8,
                     group='data_vars',
                     units='%',
                     ),
}

vec_sysdata = {
    'time': _VarAtts(dims=[],
                     dtype=np.float64,
                     group='coords',
                     default_val=nan,
                     units='',
                     ),
    'batt': _VarAtts(dims=[],
                     dtype=np.float32,
                     group='data_vars',
                     default_val=nan,
                     factor=0.1,
                     units='V',
                     ),
    'c_sound': _VarAtts(dims=[],
                        dtype=np.float32,
                        group='data_vars',
                        default_val=nan,
                        factor=0.1,
                        units='m/s',
                        ),
    'heading': _VarAtts(dims=[],
                        dtype=np.float32,
                        group='data_vars',
                        default_val=nan,
                        factor=0.1,
                        units='deg',
                        ),
    'pitch': _VarAtts(dims=[],
                      dtype=np.float32,
                      group='data_vars',
                      default_val=nan,
                      factor=0.1,
                      units='deg',
                      ),
    'roll': _VarAtts(dims=[],
                     dtype=np.float32,
                     group='data_vars',
                     default_val=nan,
                     factor=0.1,
                     units='deg',
                     ),
    'temp': _VarAtts(dims=[],
                     dtype=np.float32,
                     group='data_vars',
                     default_val=nan,
                     factor=0.01,
                     units='deg C',
                     ),
    'error': _VarAtts(dims=[],
                      dtype=np.uint8,
                      group='data_vars',
                      default_val=nan,
                      units='',
                      ),
    'status': _VarAtts(dims=[],
                       dtype=np.uint8,
                       group='data_vars',
                       default_val=nan,
                       units='',
                       ),
    'AnaIn': _VarAtts(dims=[],
                      dtype=np.float32,
                      group='sys',
                      default_val=nan,
                      units='',
                      ),
}

awac_profile = {
    'time': _VarAtts(dims=[],
                     dtype=np.float64,
                     group='coords',
                     units='',
                     ),
    'error': _VarAtts(dims=[],
                      dtype=np.uint16,
                      group='data_vars',
                      units='',
                      ),
    'AnaIn1': _VarAtts(dims=[],
                       dtype=np.float32,
                       group='sys',
                       default_val=nan,
                       units='n/a',
                       ),
    'batt': _VarAtts(dims=[],
                     dtype=np.float32,
                     group='data_vars',
                     default_val=nan,
                     factor=0.1,
                     units='V',
                     ),
    'c_sound': _VarAtts(dims=[],
                        dtype=np.float32,
                        group='data_vars',
                        default_val=nan,
                        factor=0.1,
                        units='m/s',
                        ),
    'heading': _VarAtts(dims=[],
                        dtype=np.float32,
                        group='data_vars',
                        default_val=nan,
                        factor=0.1,
                        units='deg',
                        ),
    'pitch': _VarAtts(dims=[],
                      dtype=np.float32,
                      group='data_vars',
                      default_val=nan,
                      factor=0.1,
                      units='deg',
                      ),
    'roll': _VarAtts(dims=[],
                     dtype=np.float32,
                     group='data_vars',
                     default_val=nan,
                     factor=0.1,
                     units='deg',
                     ),
    'pressure': _VarAtts(dims=[],
                         dtype=np.float32,
                         group='data_vars',
                         default_val=nan,
                         factor=0.001,
                         units='dbar',
                         ),
    'status': _VarAtts(dims=[],
                       dtype=np.float32,
                       group='data_vars',
                       default_val=nan,
                       units='',
                       ),
    'temp': _VarAtts(dims=[],
                     dtype=np.float32,
                     group='data_vars',
                     default_val=nan,
                     factor=0.01,
                     units='deg C',
                     ),
    'vel': _VarAtts(dims=[3, 'nbins', 'n'],  # how to change this for different # of beams?
                    dtype=np.float32,
                    group='data_vars',
                    default_val=nan,
                    factor=0.001,
                    units='m/s',
                    ),
    'amp': _VarAtts(dims=[3, 'nbins', 'n'],
                    dtype=np.uint8,
                    group='data_vars',
                    units='counts',
                    ),
}



import numpy as np
from copy import copy
from struct import Struct
from . import nortek2_lib as lib

dt32 = 'float32'
dt64 = 'float64'

grav = 9.81
# The starting value for the checksum:
cs0 = int('0xb58c', 0)


def _nans(*args, **kwargs):
    out = np.empty(*args, **kwargs)
    if out.dtype.kind == 'f':
        out[:] = np.NaN
    else:
        out[:] = 0
    return out


def _format(form, N):
    out = ''
    for f, n in zip(form, N):
        if n > 1:
            out += '{}'.format(n)
        out += f
    return out


class _DataDef():
    def __init__(self, list_of_defs):
        self._names = []
        self._format = []
        self._shape = []
        self._sci_func = []
        self._units = []
        self._N = []
        for itm in list_of_defs:
            self._names.append(itm[0])
            self._format.append(itm[1])
            self._shape.append(itm[2])
            self._sci_func.append(itm[3])
            if len(itm) == 5:
                self._units.append(itm[4])
            else:
                self._units.append('')
            if itm[2] == []:
                self._N.append(1)
            else:
                self._N.append(int(np.prod(itm[2])))
        self._struct = Struct('<' + self.format)
        self.nbyte = self._struct.size
        self._cs_struct = Struct('<' + '{}H'.format(int(self.nbyte // 2)))

    def init_data(self, npings):
        out = {}
        for nm, fmt, shp in zip(self._names, self._format, self._shape):
            # fmt[0] uses only the first format specifier
            # (ie, skip '15x' in 'B15x')
            out[nm] = _nans(shp + [npings], dtype=np.dtype(fmt[0]))
        return out

    def read_into(self, fobj, data, ens, cs=None):
        dat_tuple = self.read(fobj, cs=cs)
        for nm, shp, d in zip(self._names, self._shape, dat_tuple):
            try:
                data[nm][..., ens] = d
            except ValueError:
                data[nm][..., ens] = np.asarray(d).reshape(shp)

    @property
    def format(self, ):
        return _format(self._format, self._N)

    def read(self, fobj, cs=None):
        bytes = fobj.read(self.nbyte)
        if len(bytes) != self.nbyte:
            raise IOError("End of file.")
        data = self._struct.unpack(bytes)
        if cs is not None:
            if cs is True:
                # if cs is True, then it should be the last value that
                # was read.
                csval = data[-1]
                off = cs0 - csval
            elif isinstance(cs, int):
                csval = cs
                off = cs0
            cs_res = sum(self._cs_struct.unpack(bytes)) + off
            if csval is not False and (cs_res % 65536) != csval:
                raise Exception('Checksum failed!')
        out = []
        c = 0
        for idx, n in enumerate(self._N):
            if n == 1:
                out.append(data[c])
            else:
                out.append(data[c:(c + n)])
            c += n
        return out

    def read2dict(self, fobj, cs=False):
        return {self._names[idx]: dat
                for idx, dat in enumerate(self.read(fobj, cs=cs))}

    def sci_data(self, data):
        for ky, func in zip(self._names,
                            self._sci_func):
            if func is None:
                continue
            data[ky] = func(data[ky])

    def data_units(self):
        units = {}
        for ky, unit in zip(self._names, self._units):
            units[ky] = unit
        return units


class _LinFunc():
    """A simple linear offset and scaling object.

    Usage:
       scale_func = _LinFunc(scale=3, offset=5)

       new_data = scale_func(old_data)

    This will do:
       new_data = (old_data + 5) * 3
    """

    def __init__(self, scale=1, offset=0, dtype=None):
        self.scale = scale
        self.offset = offset
        self.dtype = dtype

    def __call__(self, array):
        if self.scale != 1 or self.offset != 0:
            array = (array + self.offset) * self.scale
        if self.dtype is not None:
            array = array.astype(self.dtype)
        return array


header = _DataDef([
    ('sync', 'B', [], None),
    ('hsz', 'B', [], None),
    ('id', 'B', [], None),
    ('fam', 'B', [], None),
    ('sz', 'H', [], None),
    ('cs', 'H', [], None),
    ('hcs', 'H', [], None),
])

_burst_hdr = [
    ('ver', 'B', [], None),
    ('DatOffset', 'B', [], None),
    ('config', 'H', [], None),
    ('SerialNum', 'I', [], None),
    ('year', 'B', [], None),
    ('month', 'B', [], None),
    ('day', 'B', [], None),
    ('hour', 'B', [], None),
    ('minute', 'B', [], None),
    ('second', 'B', [], None),
    ('usec100', 'H', [], None),
    ('c_sound', 'H', [], _LinFunc(0.1, dtype=dt32), 'm/s'),
    ('temp', 'H', [], _LinFunc(0.01, dtype=dt32), 'deg C'),
    ('pressure', 'I', [], _LinFunc(0.001, dtype=dt32), 'dbar'),
    ('heading', 'H', [], _LinFunc(0.01, dtype=dt32), 'deg'),
    ('pitch', 'h', [], _LinFunc(0.01, dtype=dt32), 'deg'),
    ('roll', 'h', [], _LinFunc(0.01, dtype=dt32), 'deg'),
    ('beam_config', 'H', [], None),
    ('cell_size', 'H', [], _LinFunc(0.001), 'm'),
    ('blank_dist', 'H', [], _LinFunc(0.01), 'm'),
    ('nominal_corr', 'B', [], None, '%'),
    ('temp_press', 'B', [], _LinFunc(0.2, -20, dtype=dt32), 'deg C'),
    ('batt', 'H', [], _LinFunc(0.1, dtype=dt32), 'V'),
    ('mag', 'h', [3], _LinFunc(0.1, dtype=dt32), 'uT'),
    ('accel', 'h', [3], _LinFunc(1. / 16384 * grav, dtype=dt32), 'm/s^2'),
    ('ambig_vel', 'h', [], _LinFunc(0.001, dtype=dt32), 'm/s'),
    ('data_desc', 'H', [], None),
    ('xmit_energy', 'H', [], None, 'dB'),
    ('vel_scale', 'b', [], None),
    ('power_level_dB', 'b', [], _LinFunc(dtype=dt32)),
    ('temp_mag', 'h', [], None),  # uncalibrated
    ('temp_clock', 'h', [], _LinFunc(0.01, dtype=dt32), 'deg C'),
    ('error', 'H', [], None),
    ('status0', 'H', [], None),
    ('status', 'I', [], None),
    ('_ensemble', 'I', [], None),
]

_bt_hdr = [
    ('ver', 'B', [], None),
    ('DatOffset', 'B', [], None),
    ('config', 'H', [], None),
    ('SerialNum', 'I', [], None),
    ('year', 'B', [], None),
    ('month', 'B', [], None),
    ('day', 'B', [], None),
    ('hour', 'B', [], None),
    ('minute', 'B', [], None),
    ('second', 'B', [], None),
    ('usec100', 'H', [], None),
    ('c_sound', 'H', [], _LinFunc(0.1, dtype=dt32), 'm/s'),
    ('temp', 'H', [], _LinFunc(0.01, dtype=dt32), 'deg C'),
    ('pressure', 'I', [], _LinFunc(0.001, dtype=dt32), 'dbar'),
    ('heading', 'H', [], _LinFunc(0.01, dtype=dt32), 'deg'),
    ('pitch', 'h', [], _LinFunc(0.01, dtype=dt32), 'deg'),
    ('roll', 'h', [], _LinFunc(0.01, dtype=dt32), 'deg'),
    ('beam_config', 'H', [], None),
    ('cell_size', 'H', [], _LinFunc(0.001), 'm'),
    ('blank_dist', 'H', [], _LinFunc(0.01), 'm'),
    ('nominal_corr', 'B', [], None, '%'),
    ('unused', 'B', [], None),
    ('batt', 'H', [], _LinFunc(0.1, dtype=dt32), 'V'),
    ('mag', 'h', [3], None, 'gauss'),
    ('accel', 'h', [3], _LinFunc(1. / 16384 * grav, dtype=dt32), 'm/s^2'),
    ('ambig_vel', 'I', [], _LinFunc(0.001, dtype=dt32), 'm/s'),
    ('data_desc', 'H', [], None),
    ('xmit_energy', 'H', [], None, 'dB'),
    ('vel_scale', 'b', [], None),
    ('power_level_dB', 'b', [], _LinFunc(dtype=dt32)),
    ('temp_mag', 'h', [], None),  # uncalibrated
    ('temp_clock', 'h', [], _LinFunc(0.01, dtype=dt32), 'deg C'),
    ('error', 'I', [], None),
    ('status', 'I', [], None, 'binary'),
    ('_ensemble', 'I', [], None),
]

_ahrs_def = [
    ('orientmat', 'f', [3, 3], None),
    ('quaternions', 'f', [4], None),
    ('angrt', 'f', [3], _LinFunc(np.pi / 180, dtype=dt32), 'rad/s'),
]


def _calc_bt_struct(config, nb):
    flags = lib._headconfig_int2dict(config, mode='bt')
    dd = copy(_bt_hdr)
    if flags['vel']:
        # units handled in Ad2cpReader.sci_data
        dd.append(('vel', 'i', [nb], None, 'm/s'))
    if flags['dist']:
        dd.append(('dist', 'i', [nb], _LinFunc(0.001, dtype=dt32), 'm'))
    if flags['fom']:
        dd.append(('fom', 'H', [nb], None))
    if flags['ahrs']:
        dd += _ahrs_def
    return _DataDef(dd)


def _calc_echo_struct(config, nc):
    flags = lib._headconfig_int2dict(config)
    dd = copy(_burst_hdr)
    dd[19] = ('blank_dist', 'H', [], _LinFunc(0.001))  # m
    if any([flags[nm] for nm in ['vel', 'amp', 'corr', 'alt', 'ast',
                                 'alt_raw', 'p_gd', 'std']]):
        raise Exception("Echosounder ping contains invalid data?")
    if flags['echo']:
        dd += [('echo', 'H', [nc], _LinFunc(0.01, dtype=dt32), 'dB')]
    if flags['ahrs']:
        dd += _ahrs_def
    return _DataDef(dd)


def _calc_burst_struct(config, nb, nc):
    flags = lib._headconfig_int2dict(config)
    dd = copy(_burst_hdr)
    if flags['echo']:
        raise Exception("Echosounder data found in velocity ping?")
    if flags['vel']:
        dd.append(('vel', 'h', [nb, nc], None, 'm/s'))
    if flags['amp']:
        dd.append(('amp', 'B', [nb, nc],
                   _LinFunc(0.5, dtype=dt32), 'dB'))
    if flags['corr']:
        dd.append(('corr', 'B', [nb, nc], None, '%'))
    if flags['alt']:
        # There may be a problem here with reading 32bit floats if
        # nb and nc are odd?
        dd += [('alt_dist', 'f', [], _LinFunc(dtype=dt32), 'm'),
               ('alt_quality', 'H', [], _LinFunc(0.01, dtype=dt32), 'dB'),
               ('alt_status', 'H', [], None)]
    if flags['ast']:
        dd += [
            ('ast_dist', 'f', [], _LinFunc(dtype=dt32), 'm'),
            ('ast_quality', 'H', [], _LinFunc(0.01, dtype=dt32), 'dB'),
            ('ast_offset_time', 'h', [], _LinFunc(0.0001, dtype=dt32), 's'),
            ('ast_pressure', 'f', [], None, 'dbar'),
            # This use of 'x' here is a hack
            ('ast_spare', 'B7x', [], None),
        ]
    if flags['alt_raw']:
        dd += [
            ('altraw_nsamp', 'I', [], None),
            ('altraw_dsamp', 'H', [], _LinFunc(0.0001, dtype=dt32), 'm'),
            ('altraw_samp', 'h', [], None),
        ]
    if flags['ahrs']:
        dd += _ahrs_def
    if flags['p_gd']:
        dd += [('percent_good', 'B', [nc], None, '%')]
    if flags['std']:
        dd += [('pitch_std', 'h', [],
                _LinFunc(0.01, dtype=dt32), 'deg'),
               ('roll_std', 'h', [],
                _LinFunc(0.01, dtype=dt32), 'deg'),
               ('heading_std', 'h', [],
                _LinFunc(0.01, dtype=dt32), 'deg'),
               ('press_std', 'h', [],
                _LinFunc(0.1, dtype=dt32), 'dbar'),
               # This use of 'x' here is a hack
               ('std_spare', 'H22x', [], None)]
    return _DataDef(dd)


"""
Note on "This use of 'x' is a hack": I'm afraid that using a larger
int size will give syncing problems (e.g. unpack('HB')
vs. unpack('BH')), and I need to read something, otherwise the
unpack order will get messed up. In the future, it'd be good to read
the size of the format, and hold that differently than self._N
(e.g. self._N2?)
"""



__version__ = '1.2.0'
__prog_name__ = 'DOLfYN'
__version_date__ = 'Dec-19-2022'


def ver2tuple(ver):
    out = []
    if '-' in __version__:
        ver, pre_rel = ver.split('-')
    else:
        pre_rel = None
    for val in ver.split('.'):
        val = int(val)
        out.append(val)
    if pre_rel is not None:
        return tuple(out + [pre_rel])
    else:
        return tuple(out)


version_info = ver2tuple(__version__)



#!/usr/bin/python

import argparse
import os
import numpy as np
import dolfyn
from dolfyn.rotate.base import orient2euler
import dolfyn.adv.api as avm

# TODO: add option to rotate into earth or principal frame (include
# principal_angle_True in output).

script_dir = os.path.dirname(__file__)

parser = argparse.ArgumentParser(
    description="""
    Perform motion correction of a Nortek Vector(.vec)
    file and save the output in earth(u: East, v: North, w: up)
    coordinates in units of m / s as a Matlab(TM)(.mat) file.
    """)
parser.add_argument(
    '-f',
    default=0.03333,
    help="""
    Specify the high-pass filter frequency that is applied to the
    acceleration prior to integrating acceleration (Accel) to get
    velocity. Default '0.03333' (Hz) = 30sec.
    """
)

# parser.add_argument(
# '-F',
# default=0.01,
# help="""
# Specify the high-pass filter frequency that is applied to the integrated
# acceleration (uacc) to remove drift that remains after integration. Default
# '0.01' (Hz) = 100sec
# """
# )

# parser.add_argument(
#    '-O',
#    default=None,
#    help="""NOTE: this option is deprecated and will be removed in
#    future releases. Use the '<filename>.userdata.json' method
#    instead.  Specify the 'orientation' configuration file (default:
#    '<filename>.orient', or 'vector.orient', in that
#    order). Cable-Head Vector probes the orientation of, and distance
#    between, the head to the body is arbitrary. This option specifies
#    the file which defines these variables. For more information on
#    how to measure these variables, take a look at the
#    'dolfyn-src-dir/examples/motion_correct_example.orient'
#    """
# )
parser.add_argument(
    '--fixed-head',
    action='store_true',
    help="""
    This specifies that the 'fixed-head' orientation/geometry should be used to
    compute head motion.
    """
)
parser.add_argument(
    '--mat',
    action='store_true',
    help="""
    Save the earth-frame motion-corrected data in Matlab format (default).
    """
)
# parser.add_argument(
# '--csv',
# action='store_true',
# help="""
# Save the earth-frame motion-corrected data in csv (comma-separated value) format.
# """
# )
parser.add_argument(
    '--nc',
    action='store_true',
    help="""
    Save the earth-frame motion-corrected data in the dolfyn-structured netCDF format.
    """
)

# parser.add_argument(
# '--out-earth',
# action='store_true',
# help="""
# This specifies that the output data should be return in an earth
# (u:East, v:North, w:up) coordinate system (default: True).
# """
# )

###########
# I removed this option because the data in a raw file is often
# noisy, which will lead to inaccurate estimates of the principal
# angle. Data should be cleaned prior to rotating into the principal
# frame.
# parser.add_argument(
# '--out-principal',
# action='store_false',
# help="""
# This specifies that the output data should be returned in a
# 'principal axes' frame (u:streamwise, v:cross-stream, w:up)
# coordinate system.
# """
# )
parser.add_argument(
    'filename',
    help="""The filename(s) of the the Nortek Vector file(s) to be
    processed(they probably end with '.vec').""",
    action='append'
)

args = parser.parse_args()

# if bool(args.out_principal) and bool(args.out_earth):
# raise Exception('--out-principal and --out-earth can not both be
# selected. You must choose one output frame.')
declin = None
if args.fixed_head:  # != bool(args.O):
    # Either args.fixed_head is True or args.O should be a string.
    #    if bool(args.O):
    #        exec(open(args.O).read())  # ROTMAT and VEC should be in this file.
    #        rmat = np.array(ROTMAT)
    #        vec = np.array(VEC)
    #        if 'DECLINATION' in vars():
    #            declin = DECLINATION
    #        del VEC, ROTMAT
    #    else:
    rmat = np.array([[1, 0, 0],
                     [0, 1, 0],
                     [0, 0, 1]], dtype=np.float32)
    vec = np.array([0, 0, -0.21])  # in meters
else:
    rmat = None
    vec = None

if not (args.mat or args.nc):
    args.mat = True

# Now loop over the specified file names:
for fnm in args.filename:

    dat = avm.read(fnm)

    if rmat is not None:
        dat.attrs['inst2head_rotmat'] = rmat
    if vec is not None:
        dat.attrs['inst2head_vec'] = vec
    if declin is not None:
        dat.attrs['declination'] = declin
    if ('inst2head_rotmat' not in dat.attrs or
            'inst2head_vec' not in dat.attrs):
        raise Exception("inst2head_rotmat or inst2head_vec not found "
                        "in dat.attrs. These must be specified by either:\n"
                        "  1) defining them in a '.userdata.json' file\n"
                        "  2) using the --fixed-head command-line argument\n")

    # Perform motion correction.
    if hasattr(dat, 'orientmat'):
        print('Performing motion correction...')
        # Perform the motion correction.
        dat = avm.correct_motion(dat, accel_filtfreq=args.f)
        # Compute pitch,roll,heading from orientmat.
        dat['pitch'], dat['roll'], dat['heading'] = orient2euler(dat.orientmat)
    else:
        print("""
        !!!--Warning--!!!: Orientation matrix('orientmat')
        not found. Motion correction cannot be performed on this file
        """)

    if args.mat:
        # Set matlab 'datenum' time.
        dt = dolfyn.time.epoch2date(dat['time'])
        dat['datenum'] = dolfyn.time.date2matlab(dt)

        outnm = fnm.rstrip('.vec').rstrip('.VEC') + '.mat'
        print('Saving to %s.' % outnm)
        # Save the data.
        dolfyn.io.api.save_mat(dat, outnm)

    if args.nc:
        outnm = fnm.rstrip('.vec').rstrip('.VEC') + '.nc'
        print('Saving to %s.' % outnm)
        # Save the data.
        dolfyn.save(dat, outnm)

    del dat
